import numpy as np

import pytest

from triage.component.catwalk.estimators.transformers import CutOff
from triage.component.catwalk.estimators.classifiers import ScaledLogisticRegression

from sklearn import linear_model

from sklearn import datasets
from sklearn import preprocessing
from sklearn.pipeline import Pipeline
from sklearn.model_selection import train_test_split


@pytest.fixture
def data():
    dataset = datasets.load_breast_cancer()
    X = dataset.data
    y = dataset.target

    X_train, X_test, y_train, y_test = train_test_split(
        X, y, test_size=0.3, random_state=12345
    )

    return {"X_train": X_train, "X_test": X_test, "y_train": y_train, "y_test": y_test}


def test_cutoff_warning():
    X_data = [1, 2, 0.5, 0.7, 100, -1, -23, 0]

    cutoff = CutOff()

    with pytest.raises(ValueError):
        cutoff.fit_transform(X_data)


def test_cutoff_transformer():
    cutoff = CutOff()

    X_data = np.array([1, 2, 0.5, 0.7, 100, -1, -23, 0]).reshape(-1,1)

    assert np.all(cutoff.fit_transform(X_data) == np.array([1, 1, 0.5, 0.7, 1, 0, 0, 0]).reshape(-1,1))


def test_cutoff_inside_a_pipeline(data):
    minmax_scaler = preprocessing.MinMaxScaler()
    dsapp_cutoff = CutOff()

    pipeline = Pipeline(
        [("minmax_scaler", minmax_scaler), ("dsapp_cutoff", dsapp_cutoff)]
    )

    pipeline.fit(data["X_train"], data["y_train"])

    X_fake_new_data = data["X_test"][-1, :].reshape(1, -1) + 0.5

    mms = preprocessing.MinMaxScaler().fit(data["X_train"])

    assert np.all(
        (mms.transform(X_fake_new_data) > 1)
        == (pipeline.transform(X_fake_new_data) == 1)
    )


def test_dsapp_lr(data):
    dsapp_lr = ScaledLogisticRegression()
    dsapp_lr.fit(data["X_train"], data["y_train"])

    minmax_scaler = preprocessing.MinMaxScaler()
    dsapp_cutoff = CutOff()
    lr = linear_model.LogisticRegression(solver='lbfgs')

    pipeline = Pipeline(
        [("minmax_scaler", minmax_scaler), ("dsapp_cutoff", dsapp_cutoff), ("lr", lr)]
    )

    pipeline.fit(data["X_train"], data["y_train"])

    assert np.all(dsapp_lr.predict(data["X_test"]) == pipeline.predict(data["X_test"]))



# -*- coding: utf-8 -*-
"""test_spacetime

Unit tests for `collate.spacetime` module.

"""
from datetime import date
from itertools import product

import pytest
import sqlalchemy
import testing.postgresql

from triage.component.collate import Aggregate, SpacetimeAggregation


events_data = [
    # entity id, event_date, outcome
    [1, date(2014, 1, 1), True],
    [1, date(2014, 11, 10), False],
    [1, date(2015, 1, 1), False],
    [1, date(2015, 11, 10), True],
    [2, date(2013, 6, 8), True],
    [2, date(2014, 6, 8), False],
    [3, date(2014, 3, 3), False],
    [3, date(2014, 7, 24), False],
    [3, date(2015, 3, 3), True],
    [3, date(2015, 7, 24), False],
    [4, date(2015, 12, 13), False],
    [4, date(2016, 12, 13), True],
]

# distinct entity_id, event_date pairs
state_data = sorted(
    list(
        product(
            set([l[0] for l in events_data]),
            set([l[1] for l in events_data] + [date(2016, 1, 1)]),
        )
    )
)


def test_basic_spacetime():
    with testing.postgresql.Postgresql() as psql:
        engine = sqlalchemy.create_engine(psql.url())
        engine.execute(
            "create table events (entity_id int, event_date date, outcome bool)"
        )
        for event in events_data:
            engine.execute("insert into events values (%s, %s, %s::bool)", event)

        engine.execute("create table states (entity_id int, as_of_date date)")
        for state in state_data:
            engine.execute("insert into states values (%s, %s)", state)

        agg = Aggregate(
            "outcome::int",
            ["sum", "avg", "stddev"],
            {
                "coltype": "aggregate",
                "avg": {"type": "mean"},
                "sum": {"type": "constant", "value": 3},
                "stddev": {"type": "constant", "value": 2},
            },
        )
        st = SpacetimeAggregation(
            aggregates=[agg],
            from_obj="events",
            groups=["entity_id"],
            intervals=["1y", "2y", "all"],
            dates=["2016-01-01", "2015-01-01"],
            state_table="states",
            state_group="entity_id",
            date_column="event_date",
            output_date_column="as_of_date",
        )

        st.execute(engine.connect())

        r = engine.execute(
            "select * from events_entity_id order by entity_id, as_of_date"
        )
        rows = [x for x in r]
        assert rows[0]["entity_id"] == 1
        assert rows[0]["as_of_date"] == date(2015, 1, 1)
        assert rows[0]["events_entity_id_1y_outcome::int_sum"] == 1
        assert rows[0]["events_entity_id_1y_outcome::int_avg"] == 0.5
        assert rows[0]["events_entity_id_2y_outcome::int_sum"] == 1
        assert rows[0]["events_entity_id_2y_outcome::int_avg"] == 0.5
        assert rows[0]["events_entity_id_all_outcome::int_sum"] == 1
        assert rows[0]["events_entity_id_all_outcome::int_avg"] == 0.5
        assert rows[1]["entity_id"] == 1
        assert rows[1]["as_of_date"] == date(2016, 1, 1)
        assert rows[1]["events_entity_id_1y_outcome::int_sum"] == 1
        assert rows[1]["events_entity_id_1y_outcome::int_avg"] == 0.5
        assert rows[1]["events_entity_id_2y_outcome::int_sum"] == 2
        assert rows[1]["events_entity_id_2y_outcome::int_avg"] == 0.5
        assert rows[1]["events_entity_id_all_outcome::int_sum"] == 2
        assert rows[1]["events_entity_id_all_outcome::int_avg"] == 0.5

        assert rows[2]["entity_id"] == 2
        assert rows[2]["as_of_date"] == date(2015, 1, 1)
        assert rows[2]["events_entity_id_1y_outcome::int_sum"] == 0
        assert rows[2]["events_entity_id_1y_outcome::int_avg"] == 0
        assert rows[2]["events_entity_id_2y_outcome::int_sum"] == 1
        assert rows[2]["events_entity_id_2y_outcome::int_avg"] == 0.5
        assert rows[2]["events_entity_id_all_outcome::int_sum"] == 1
        assert rows[2]["events_entity_id_all_outcome::int_avg"] == 0.5
        assert rows[3]["entity_id"] == 2
        assert rows[3]["as_of_date"] == date(2016, 1, 1)
        assert rows[3]["events_entity_id_1y_outcome::int_sum"] is None
        assert rows[3]["events_entity_id_1y_outcome::int_avg"] is None
        assert rows[3]["events_entity_id_2y_outcome::int_sum"] == 0
        assert rows[3]["events_entity_id_2y_outcome::int_avg"] == 0
        assert rows[3]["events_entity_id_all_outcome::int_sum"] == 1
        assert rows[3]["events_entity_id_all_outcome::int_avg"] == 0.5

        assert rows[4]["entity_id"] == 3
        assert rows[4]["as_of_date"] == date(2015, 1, 1)
        assert rows[4]["events_entity_id_1y_outcome::int_sum"] == 0
        assert rows[4]["events_entity_id_1y_outcome::int_avg"] == 0
        assert rows[4]["events_entity_id_2y_outcome::int_sum"] == 0
        assert rows[4]["events_entity_id_2y_outcome::int_avg"] == 0
        assert rows[4]["events_entity_id_all_outcome::int_sum"] == 0
        assert rows[4]["events_entity_id_all_outcome::int_avg"] == 0
        assert rows[5]["entity_id"] == 3
        assert rows[5]["as_of_date"] == date(2016, 1, 1)
        assert rows[5]["events_entity_id_1y_outcome::int_sum"] == 1
        assert rows[5]["events_entity_id_1y_outcome::int_avg"] == 0.5
        assert rows[5]["events_entity_id_2y_outcome::int_sum"] == 1
        assert rows[5]["events_entity_id_2y_outcome::int_avg"] == 0.25
        assert rows[5]["events_entity_id_all_outcome::int_sum"] == 1
        assert rows[5]["events_entity_id_all_outcome::int_avg"] == 0.25

        assert rows[6]["entity_id"] == 4
        # rows[6]['date'] == date(2015, 1, 1) is skipped due to no data!
        assert rows[6]["as_of_date"] == date(2016, 1, 1)
        assert rows[6]["events_entity_id_1y_outcome::int_sum"] == 0
        assert rows[6]["events_entity_id_1y_outcome::int_avg"] == 0
        assert rows[6]["events_entity_id_2y_outcome::int_sum"] == 0
        assert rows[6]["events_entity_id_2y_outcome::int_avg"] == 0
        assert rows[6]["events_entity_id_all_outcome::int_sum"] == 0
        assert rows[6]["events_entity_id_all_outcome::int_avg"] == 0
        assert len(rows) == 7

        # check some imputation results
        r = engine.execute(
            "select * from events_aggregation_imputed order by entity_id, as_of_date"
        )
        rows = [x for x in r]
        assert rows[6]["entity_id"] == 4
        assert rows[6]["as_of_date"] == date(2015, 1, 1)
        assert rows[6]["events_entity_id_1y_outcome::int_sum"] == 3
        assert rows[6]["events_entity_id_1y_outcome::int_imp"] == 1
        assert rows[6]["events_entity_id_1y_outcome::int_stddev"] == 2
        assert rows[6]["events_entity_id_1y_outcome::int_stddev_imp"] == 1
        assert (
            round(float(rows[6]["events_entity_id_1y_outcome::int_avg"]), 4) == 0.1667
        )
        assert rows[6]["events_entity_id_2y_outcome::int_sum"] == 3
        assert rows[6]["events_entity_id_2y_outcome::int_imp"] == 1
        assert rows[6]["events_entity_id_2y_outcome::int_stddev"] == 2
        assert rows[6]["events_entity_id_2y_outcome::int_stddev_imp"] == 1
        assert (
            round(float(rows[6]["events_entity_id_2y_outcome::int_avg"]), 4) == 0.3333
        )
        assert rows[6]["events_entity_id_all_outcome::int_sum"] == 3
        assert rows[6]["events_entity_id_all_outcome::int_imp"] == 1
        assert rows[6]["events_entity_id_all_outcome::int_stddev"] == 2
        assert rows[6]["events_entity_id_all_outcome::int_stddev_imp"] == 1
        assert (
            round(float(rows[6]["events_entity_id_all_outcome::int_avg"]), 4) == 0.3333
        )
        assert rows[6]["events_entity_id_all_outcome::int_imp"] == 1
        assert rows[7]["entity_id"] == 4
        assert rows[7]["as_of_date"] == date(2016, 1, 1)
        assert rows[7]["events_entity_id_1y_outcome::int_sum"] == 0
        assert rows[7]["events_entity_id_1y_outcome::int_imp"] == 0
        assert rows[7]["events_entity_id_1y_outcome::int_avg"] == 0
        assert rows[7]["events_entity_id_1y_outcome::int_stddev"] == 2
        assert rows[7]["events_entity_id_1y_outcome::int_stddev_imp"] == 1
        assert rows[7]["events_entity_id_2y_outcome::int_sum"] == 0
        assert rows[7]["events_entity_id_2y_outcome::int_imp"] == 0
        assert rows[7]["events_entity_id_2y_outcome::int_avg"] == 0
        assert rows[7]["events_entity_id_2y_outcome::int_stddev"] == 2
        assert rows[7]["events_entity_id_2y_outcome::int_stddev_imp"] == 1
        assert rows[7]["events_entity_id_all_outcome::int_sum"] == 0
        assert rows[7]["events_entity_id_all_outcome::int_imp"] == 0
        assert rows[7]["events_entity_id_all_outcome::int_avg"] == 0
        assert rows[7]["events_entity_id_all_outcome::int_stddev"] == 2
        assert rows[7]["events_entity_id_all_outcome::int_stddev_imp"] == 1
        assert len(rows) == 8


def test_input_min_date():
    with testing.postgresql.Postgresql() as psql:
        engine = sqlalchemy.create_engine(psql.url())
        engine.execute("create table events (entity_id int, date date, outcome bool)")
        for event in events_data:
            engine.execute("insert into events values (%s, %s, %s::bool)", event)

        engine.execute("create table states (entity_id int, date date)")
        for state in state_data:
            engine.execute("insert into states values (%s, %s)", state)

        agg = Aggregate(
            "outcome::int",
            ["sum", "avg"],
            {
                "coltype": "aggregate",
                "avg": {"type": "mean"},
                "sum": {"type": "constant", "value": 3},
                "max": {"type": "zero"},
            },
        )
        st = SpacetimeAggregation(
            aggregates=[agg],
            from_obj="events",
            groups=["entity_id"],
            intervals=["all"],
            dates=["2016-01-01"],
            state_table="states",
            state_group="entity_id",
            date_column='"date"',
            input_min_date="2015-11-10",
        )

        st.execute(engine.connect())

        r = engine.execute("select * from events_entity_id order by entity_id")
        rows = [x for x in r]

        assert rows[0]["entity_id"] == 1
        assert rows[0]["date"] == date(2016, 1, 1)
        assert rows[0]["events_entity_id_all_outcome::int_sum"] == 1
        assert rows[0]["events_entity_id_all_outcome::int_avg"] == 1
        assert rows[1]["entity_id"] == 4
        assert rows[1]["date"] == date(2016, 1, 1)
        assert rows[1]["events_entity_id_all_outcome::int_sum"] == 0
        assert rows[1]["events_entity_id_all_outcome::int_avg"] == 0

        assert len(rows) == 2

        st = SpacetimeAggregation(
            aggregates=[agg],
            from_obj="events",
            groups=["entity_id"],
            intervals=["1y", "all"],
            dates=["2016-01-01", "2015-01-01"],
            state_table="states",
            state_group="entity_id",
            date_column='"date"',
            input_min_date="2014-11-10",
        )
        with pytest.raises(ValueError):
            st.validate(engine.connect())
        with pytest.raises(ValueError):
            st.execute(engine.connect())


def test_join_with_cohort_table(db_engine):
    # if we specify joining with the cohort table
    # only entity_id/date pairs in the cohort table should show up
    db_engine.execute("create table events (entity_id int, date date, outcome bool)")
    for event in events_data:
        db_engine.execute("insert into events values (%s, %s, %s::bool)", event)

    db_engine.execute("create table cohort (entity_id int, date date)")

    # use the states list from above except only include entities 1 and 2 in the cohort
    smaller_cohort = sorted(
        product(
            set([l[0] for l in events_data if l[0] == 1 or l[0] == 2]),
            set([l[1] for l in events_data] + [date(2016, 1, 1)]),
        )
    )
    for state in smaller_cohort:
        db_engine.execute("insert into cohort values (%s, %s)", state)

    # create our test aggregation with the important 'join_with_cohort_table' flag
    agg = Aggregate(
        "outcome::int",
        ["sum", "avg"],
        {
            "coltype": "aggregate",
            "avg": {"type": "mean"},
            "sum": {"type": "constant", "value": 3},
            "max": {"type": "zero"},
        },
    )
    st = SpacetimeAggregation(
        aggregates=[agg],
        from_obj="events",
        groups=["entity_id"],
        intervals=["all"],
        dates=["2016-01-01", "2015-01-01"],
        state_table="cohort",
        state_group="entity_id",
        date_column='"date"',
        join_with_cohort_table=True,
    )

    st.execute(db_engine.connect())

    r = db_engine.execute("select * from events_entity_id order by entity_id, date")
    rows = [x for x in r]

    # these rows should be similar to the rows in the basic spacetime test,
    # except only the rows for entities 1 and 2 are present
    assert len(rows) == 4

    assert rows[0]["entity_id"] == 1
    assert rows[0]["date"] == date(2015, 1, 1)
    assert rows[0]["events_entity_id_all_outcome::int_sum"] == 1
    assert rows[0]["events_entity_id_all_outcome::int_avg"] == 0.5
    assert rows[1]["entity_id"] == 1
    assert rows[1]["date"] == date(2016, 1, 1)
    assert rows[1]["events_entity_id_all_outcome::int_sum"] == 2
    assert rows[1]["events_entity_id_all_outcome::int_avg"] == 0.5

    assert rows[2]["entity_id"] == 2
    assert rows[2]["date"] == date(2015, 1, 1)
    assert rows[2]["events_entity_id_all_outcome::int_sum"] == 1
    assert rows[2]["events_entity_id_all_outcome::int_avg"] == 0.5
    assert rows[3]["entity_id"] == 2
    assert rows[3]["date"] == date(2016, 1, 1)
    assert rows[3]["events_entity_id_all_outcome::int_sum"] == 1
    assert rows[3]["events_entity_id_all_outcome::int_avg"] == 0.5



import verboselogs, logging
logger = verboselogs.VerboseLogger(__name__)

import math

import numpy as np
from sqlalchemy.orm import sessionmaker
from sqlalchemy import or_

from .utils import db_retry, retrieve_model_hash_from_id, save_db_objects, sort_predictions_and_labels, AVAILABLE_TIEBREAKERS
from triage.component.results_schema import Model
from triage.util.db import scoped_session
from triage.util.random import generate_python_random_seed
import ohio.ext.pandas
import pandas as pd


class Predictor:
    expected_matrix_ts_format = "%Y-%m-%d %H:%M:%S"
    available_tiebreakers = AVAILABLE_TIEBREAKERS

    def __init__(
        self,
        model_storage_engine,
        db_engine,
        rank_order,
        replace=True,
        save_predictions=True
    ):
        """Encapsulates the task of generating predictions on an arbitrary
        dataset and storing the results

        Args:
            model_storage_engine (catwalk.storage.ModelStorageEngine)
            db_engine (sqlalchemy.engine)
            rank_order

        """
        self.model_storage_engine = model_storage_engine
        self.db_engine = db_engine
        self.rank_order = rank_order
        self.replace = replace
        self.save_predictions = save_predictions

    @property
    def sessionmaker(self):
        return sessionmaker(bind=self.db_engine)

    @db_retry
    def load_model(self, model_id):
        """Downloads the cached model associated with a given model id

        Args:
            model_id (int) The id of a given model in the database

        Returns:
            A python object which implements .predict()
        """

        model_hash = retrieve_model_hash_from_id(self.db_engine, model_id)
        logger.spam(f"Checking for model_hash {model_hash} in store")
        if self.model_storage_engine.exists(model_hash):
            return self.model_storage_engine.load(model_hash)

    @db_retry
    def delete_model(self, model_id):
        """Deletes the cached model associated with a given model id

        Args:
            model_id (int) The id of a given model in the database
        """
        model_hash = retrieve_model_hash_from_id(self.db_engine, model_id)
        self.model_storage_engine.delete(model_hash)

    @db_retry
    def _existing_predictions(self, Prediction_obj, session, model_id, matrix_store):
        logger.debug(f"Looking for existing predictions for model {model_id} on {matrix_store.matrix_type.string_name} matrix [{matrix_store.uuid}]")
        return (
            session.query(Prediction_obj)
            .filter_by(model_id=model_id)
            .filter(Prediction_obj.as_of_date.in_(matrix_store.as_of_dates))
        )

    @db_retry
    def needs_predictions(self, matrix_store, model_id):
        """Returns whether or not the given matrix and model are lacking any predictions

        Args:
            matrix_store (triage.component.catwalk.storage.MatrixStore) A matrix with metadata
            model_id (int) A database ID of a model

        The way we check is by grabbing all the distinct as-of-dates in the predictions table
        for this model and matrix. If there are more as-of-dates defined in the matrix's metadata
        than are in the table, we need predictions
        """
        if not self.save_predictions:
            return False
        session = self.sessionmaker()
        prediction_obj = matrix_store.matrix_type.prediction_obj
        as_of_dates_in_db = set(
            as_of_date.date()
            for (as_of_date,) in session.query(prediction_obj).filter_by(
                model_id=model_id,
                matrix_uuid=matrix_store.uuid
            ).distinct(prediction_obj.as_of_date).values("as_of_date")
        )
        as_of_dates_needed = set(matrix_store.as_of_dates)
        needed = bool(as_of_dates_needed - as_of_dates_in_db)
        session.close()
        return needed

    @db_retry
    def _load_saved_predictions(self, existing_predictions, matrix_store):
        index = matrix_store.index
        score_lookup = {}
        for prediction in existing_predictions:
            score_lookup[
                (prediction.entity_id, prediction.as_of_date.date())
            ] = prediction.score
        score_iterator = (
            score_lookup[(entity_id, dt.date())] for (entity_id, dt) in index
        )
        return np.fromiter(score_iterator, float)

    @db_retry
    def _write_predictions_to_db(
        self,
        model_id,
        matrix_store,
        df,
        misc_db_parameters,
        Prediction_obj,
    ):
        """Writes given predictions to database

        entity_ids, predictions, labels are expected to be in the same order

        Args:
            model_id (int) the id of the model associated with the given predictions
            matrix_store (catwalk.storage.MatrixStore) the matrix and metadata
            df (pd.DataFrame) with the following columns entity_id, as_of_date, score, label_value and rank_abs_no_ties, rank_abs_with_ties, rank_pct_no_ties, rank_pct_with_ties
    predictions (iterable) predicted values
            Prediction_obj (TrainPrediction or TestPrediction) table to store predictions to

        """
        try:
            session = self.sessionmaker()
            existing_predictions = self._existing_predictions(
                Prediction_obj, session, model_id, matrix_store
            )
            if existing_predictions.count() > 0:
                existing_predictions.delete(synchronize_session=False)
                logger.info(f"Found old predictions for model {model_id} on {matrix_store.matrix_type.string_name} matrix {matrix_store.uuid}. Those predictions were deleted.")
            session.expire_all()
            session.commit()
        finally:
            session.close()
        test_label_timespan = matrix_store.metadata["label_timespan"]

        record_stream = (
            Prediction_obj(
                model_id=int(model_id),
                entity_id=int(row.entity_id),
                as_of_date=row.as_of_date,
                score=float(row.score),
                label_value=int(row.label_value) if not math.isnan(row.label_value) else None,
                rank_abs_no_ties = int(row.rank_abs_no_ties),
                rank_abs_with_ties = int(row.rank_abs_with_ties),
                rank_pct_no_ties = row.rank_pct_no_ties,
                rank_pct_with_ties = row.rank_pct_with_ties,
                matrix_uuid=matrix_store.uuid,
                test_label_timespan=test_label_timespan,
                **misc_db_parameters
            ) for row in df.itertuples()
        )
        save_db_objects(self.db_engine, record_stream)

    def _write_metadata_to_db(self, model_id, matrix_uuid, matrix_type, random_seed):
        orm_obj = matrix_type.prediction_metadata_obj(
            model_id=model_id,
            matrix_uuid=matrix_uuid,
            tiebreaker_ordering=self.rank_order,
            random_seed=random_seed,
            predictions_saved=self.save_predictions,
        )
        session = self.sessionmaker()
        session.merge(orm_obj)
        session.commit()
        session.close()

    def _needs_ranks(self, model_id, matrix_uuid, matrix_type):
        if self.replace:
            logger.info("Replace flag set, will compute and store ranks regardless")
            return True
        with scoped_session(self.db_engine) as session:
            # if the metadata is different (e.g. they changed the rank order)
            # or there are any null ranks we need to rank
            metadata_matches = session.query(session.query(matrix_type.prediction_metadata_obj).filter_by(
                model_id=model_id,
                matrix_uuid=matrix_uuid,
                tiebreaker_ordering=self.rank_order,
            ).exists()).scalar()
            if not metadata_matches:
                logger.debug("Prediction metadata does not match what is in configuration"
                              ", will compute and store ranks")
                return True

            any_nulls_in_ranks = session.query(session.query(matrix_type.prediction_obj)\
                .filter(
                    matrix_type.prediction_obj.model_id == model_id,
                    matrix_type.prediction_obj.matrix_uuid == matrix_uuid,
                    or_(
                        matrix_type.prediction_obj.rank_abs_no_ties == None,
                        matrix_type.prediction_obj.rank_abs_with_ties == None,
                        matrix_type.prediction_obj.rank_pct_no_ties == None,
                        matrix_type.prediction_obj.rank_pct_with_ties == None,
                    )
                ).exists()).scalar()
            if any_nulls_in_ranks:
                logger.debug("At least one null in rankings in predictions table",
                              ", will compute and store ranks")
                return True
        logger.debug("No need to recompute prediction ranks")
        return False


    def predict(self, model_id, matrix_store, misc_db_parameters, train_matrix_columns):
        """Generate predictions and store them in the database

        Args:
            model_id (int) the id of the trained model to predict based off of
            matrix_store (catwalk.storage.MatrixStore) a wrapper for the
                prediction matrix and metadata
            misc_db_parameters (dict): attributes and values to add to each
                TrainPrediction or TestPrediction object in the results schema
            train_matrix_columns (list): The order of columns that the model
                was trained on

        Returns:
            (np.Array) the generated prediction values
        """
        # Setting the Prediction object type - TrainPrediction or TestPrediction
        matrix_type = matrix_store.matrix_type

        if not self.replace:
            logger.info(
                f"Replace flag not set, looking for old predictions for model id {model_id} "
                f"on {matrix_store.matrix_type.string_name} matrix {matrix_store.uuid}"
            )
            try:
                session = self.sessionmaker()
                existing_predictions = self._existing_predictions(
                    matrix_type.prediction_obj, session, model_id, matrix_store
                )
                logger.spam(f"Existing predictions length: {existing_predictions.count()}, Length of matrix: {len(matrix_store.index)}")
                if existing_predictions.count() == len(matrix_store.index):
                    logger.info(
                        f"Found old predictions for model id {model_id}, matrix {matrix_store.uuid}, returning saved versions"
                    )
                    return self._load_saved_predictions(existing_predictions, matrix_store)
            finally:
                session.close()

        model = self.load_model(model_id)
        if not model:
            raise ValueError(f"Model id {model_id} not found")
        logger.spam(f"Loaded model {model_id}")

        # Labels are popped from matrix (i.e. they are removed and returned)
        labels = matrix_store.labels

        predictions = model.predict_proba(
            matrix_store.matrix_with_sorted_columns(train_matrix_columns)
        )[:, 1]  # Returning only the scores for the label == 1


        logger.debug(
            f"Generated predictions for model {model_id} on {matrix_store.matrix_type.string_name} matrix {matrix_store.uuid}"
        )
        if self.save_predictions:
            df = pd.DataFrame(data=None, columns=None, index=matrix_store.index)
            df['label_value'] = matrix_store.labels
            df['score'] = predictions


            logger.spam(f"Sorting predictions for model {model_id} using {self.rank_order}")

            if self.rank_order == 'best':
                df.sort_values(by=["score", "label_value"], inplace=True, ascending=[False,False], na_position='last')
            elif self.rank_order == 'worst':
                df.sort_values(by=["score", "label_value"], inplace=True, ascending=[False,True], na_position='first')
            elif self.rank_order == 'random':
                df['random'] = np.random.rand(len(df))
                df.sort_values(by=['score', 'random'], inplace=True, ascending=[False, False])
                df.drop('random', axis=1)
            else:
                raise ValueError(f"Rank order specified in condiguration file not recognized: {self.rank_order} ")

            df['rank_abs_no_ties'] = df['score'].rank(ascending=False, method='first')
            # uses the lowest rank in the group
            df['rank_abs_with_ties'] = df['score'].rank(ascending=False, method='min')
            # No gaps between groups (so it reaches 1.0). We are using rank_abs_no_ties so we can
            # respect that order (instead of using the mathematical formula,  as was done before)
            df['rank_pct_no_ties'] = df['rank_abs_no_ties'].rank(ascending=True, method='dense', pct=True)
            df['rank_pct_with_ties'] = df['score'].rank(ascending=False, method='dense', pct=True)

            df.reset_index(inplace=True)
            logger.debug(f"Predictions on {matrix_store.matrix_type.string_name} matrix {matrix_store.uuid} from model {model_id} sorted using {self.rank_order}")

            logger.spam(
                f"Writing predictions for model {model_id} on {matrix_store.matrix_type.string_name}  matrix {matrix_store.uuid} to database"
            )

            self._write_predictions_to_db(
                model_id,
                matrix_store,
                df,
                misc_db_parameters,
                matrix_type.prediction_obj,
            )
            logger.debug(
                f"Wrote predictions for model {model_id} on  {matrix_store.matrix_type.string_name} matrix {matrix_store.uuid} to database"
            )
        else:
            logger.notice(
                f"Predictions for model {model_id} on {matrix_store.matrix_type.string_name} matrix {matrix_store.uuid}  weren't written to the db because, because you asked not to do so"
            )
            logger.spam(f"Status of the save_predictions flag: {self.save_predictions}")

        self._write_metadata_to_db(
            model_id=model_id,
            matrix_uuid=matrix_store.uuid,
            matrix_type=matrix_type,
            random_seed=None,
        )

        return predictions



import copy
import datetime
import importlib

import verboselogs, logging
logger = verboselogs.VerboseLogger(__name__)

import random
import sys
from contextlib import contextmanager

import numpy as np
import pandas as pd

from sklearn.model_selection import ParameterGrid
from sqlalchemy.orm import sessionmaker

from triage.util.random import generate_python_random_seed
from triage.component.results_schema import Model, FeatureImportance
from triage.component.catwalk.exceptions import BaselineFeatureNotInMatrix
from triage.tracking import built_model, skipped_model, errored_model

from .model_grouping import ModelGrouper
from .feature_importances import get_feature_importances
from .utils import (
    filename_friendly_hash,
    retrieve_model_id_from_hash,
    db_retry,
    save_db_objects,
)

NO_FEATURE_IMPORTANCE = (
    "Algorithm does not support a standard way" + " to calculate feature importance."
)


def flatten_grid_config(grid_config):
    """Flattens a model/parameter grid configuration into individually
    trainable model/parameter pairs

    Yields: (tuple) classpath and parameters
    """
    for class_path, parameter_config in grid_config.items():
        for parameters in ParameterGrid(parameter_config):
            yield class_path, parameters


class ModelTrainer:
    """Trains a series of classifiers using the same training set
    Args:
        project_path (string) path to project folder,
            under which to cache model pickles
        experiment_hash (string) foreign key to the triage_metadata.experiments table
        model_storage_engine (catwalk.storage.ModelStorageEngine)
        db_engine (sqlalchemy.engine)
        replace (bool) whether or not to replace existing versions of models
    """

    def __init__(
        self,
        experiment_hash,
        model_storage_engine,
        db_engine,
        model_grouper=None,
        replace=True,
        run_id=None,
    ):
        self.experiment_hash = experiment_hash
        self.model_storage_engine = model_storage_engine
        self.model_grouper = model_grouper or ModelGrouper()
        self.db_engine = db_engine
        self.replace = replace
        self.run_id = run_id

    @property
    def sessionmaker(self):
        return sessionmaker(bind=self.db_engine)

    def unique_parameters(self, parameters):
        return {key: parameters[key] for key in parameters.keys() if key != "n_jobs"}

    def _model_hash(self, matrix_metadata, class_path, parameters, random_seed):
        """Generates a unique identifier for a trained model
        based on attributes of the model that together define
        equivalence; in other words, if we train a second model with these
        same attributes there would be no reason to keep the old one)

        Args:
        class_path (string): a full class path for the classifier
        parameters (dict): all hyperparameters to be passed to the classifier
        random_seed (int) an integer suitable for seeding the random generator before training

        Returns: (string) a unique identifier
        """

        unique = {
            "className": class_path,
            "parameters": self.unique_parameters(parameters),
            "project_path": self.model_storage_engine.project_storage.project_path,
            "training_metadata": matrix_metadata,
            "random_seed": random_seed,
        }
        logger.spam(f"Creating model hash from unique data {unique}")
        return filename_friendly_hash(unique)

    def _train(self, matrix_store, class_path, parameters):
        """Fit a model to a training set. Works on any modeling class that
        is available in this package's environment and implements .fit

        Args:
            class_path (string) A full classpath to the model class
            parameters (dict) hyperparameters to give to the model constructor

        Returns:
            tuple of (fitted model, list of column names without label)
        """
        module_name, class_name = class_path.rsplit(".", 1)
        module = importlib.import_module(module_name)
        cls = getattr(module, class_name)
        instance = cls(**parameters)

        return instance.fit(matrix_store.design_matrix, matrix_store.labels)

    @db_retry
    def _save_feature_importances(self, model_id, feature_importances, feature_names):
        """Saves feature importances to the database.

        Deletes any existing feature importances for the given model_id.

        Args:
            model_id (int) The database id for the model
            feature_importances (np.ndarray, maybe). Calculated feature importances
                for the model
            feature_names (list) Feature names for the corresponding entries in feature_importances
        """
        self.db_engine.execute(
            "delete from train_results.feature_importances where model_id = %s",
            model_id,
        )
        db_objects = []
        if isinstance(feature_importances, np.ndarray):
            temp_df = pd.DataFrame({"feature_importance": feature_importances})
            features_index = temp_df.index.tolist()
            rankings_abs = temp_df["feature_importance"].rank(
                method="dense", ascending=False
            )
            rankings_pct = temp_df["feature_importance"].rank(
                method="dense", ascending=False, pct=True
            )
            for feature_index, importance, rank_abs, rank_pct in zip(
                features_index, feature_importances, rankings_abs, rankings_pct
            ):
                db_objects.append(
                    FeatureImportance(
                        model_id=model_id,
                        feature_importance=round(float(importance), 10),
                        feature=feature_names[feature_index],
                        rank_abs=int(rank_abs),
                        rank_pct=round(float(rank_pct), 10),
                    )
                )
        # get_feature_importances was not able to find
        # feature importances
        else:
            db_objects.append(
                FeatureImportance(
                    model_id=model_id,
                    feature_importance=0,
                    feature=NO_FEATURE_IMPORTANCE,
                    rank_abs=0,
                    rank_pct=0,
                )
            )
        save_db_objects(self.db_engine, db_objects)

    @db_retry
    def _write_model_to_db(
        self,
        class_path,
        parameters,
        feature_names,
        model_hash,
        trained_model,
        model_group_id,
        model_size,
        misc_db_parameters,
    ):
        """Writes model and feature importance data to a database
        Will overwrite the data of any previous versions
        (any existing model that shares a hash)

        If the replace flag on the object is set, the existing version of the model
        will have its non-unique attributes (e.g. timestamps) updated,
        and feature importances fully replaced.

        If the replace flag on the object is not set, the existing model metadata
        and feature importances will be used.

        Args:
            class_path (string) A full classpath to the model class
            parameters (dict) hyperparameters to give to the model constructor
            feature_names (list) feature names in order given to model
            model_hash (string) a unique id for the model
            trained_model (object) a trained model object
            model_group_id (int) the unique id for the model group
            model_size (float) the size of the stored model in kB
            misc_db_parameters (dict) params to pass through to the database
        """
        model_id = retrieve_model_id_from_hash(self.db_engine, model_hash)
        if model_id and not self.replace:
            logger.notice(
                f"Metadata for model {model_id} found in database. Reusing model metadata."
            )
            return model_id
        else:
            model = Model(
                model_hash=model_hash,
                model_type=class_path,
                hyperparameters=parameters,
                model_group_id=model_group_id,
                built_by_experiment=self.experiment_hash,
                built_in_experiment_run=self.run_id,
                model_size=model_size,
                **misc_db_parameters,
            )
            session = self.sessionmaker()
            if model_id:
                logger.notice(
                    f"Found model {model_id}, updating non-unique attributes"
                )
                model.model_id = model_id
                session.merge(model)
                session.commit()
            else:
                session.add(model)
                session.commit()
                model_id = model.model_id
                logger.notice(f"Model {model_id}, not found from previous runs. Adding the new model")
            session.close()

        logger.spam(f"Saving feature importances for model_id {model_id}")
        self._save_feature_importances(
            model_id, get_feature_importances(trained_model), feature_names
        )
        logger.debug(f"Saved feature importances for model_id {model_id}")
        return model_id

    def _train_and_store_model(
        self, matrix_store, class_path, parameters, model_hash, misc_db_parameters, random_seed
    ):
        """Train a model, cache it, and write metadata to a database

        Args:
            matrix_store(catwalk.storage.MatrixStore) a matrix and metadata
            class_path (string) A full classpath to the model class
            parameters (dict) hyperparameters to give to the model constructor
            model_hash (string) a unique id for the model
            misc_db_parameters (dict) params to pass through to the database

        Returns: (int) a database id for the model
        """
        random.seed(random_seed)
        misc_db_parameters["random_seed"] = random_seed
        misc_db_parameters["run_time"] = datetime.datetime.now().isoformat()
        logger.debug(f"Training and storing model for matrix uuid {matrix_store.uuid}")
        trained_model = self._train(matrix_store, class_path, parameters)

        unique_parameters = self.unique_parameters(parameters)

        model_group_id = self.model_grouper.get_model_group_id(
            class_path, unique_parameters, matrix_store.metadata, self.db_engine
        )
        logger.debug(
            f"Trained model: hash {model_hash}, model group {model_group_id} "
        )
        # Writing th model to storage, then getting its size in kilobytes.
        self.model_storage_engine.write(trained_model, model_hash)
        model_size = sys.getsizeof(trained_model) / (1024.0)

        logger.spam(f"Cached model: {model_hash}")
        model_id = self._write_model_to_db(
            class_path,
            unique_parameters,
            matrix_store.columns(include_label=False),
            model_hash,
            trained_model,
            model_group_id,
            model_size,
            misc_db_parameters,
        )
        logger.debug(f"Wrote model {model_id} [{model_hash}] to db")
        return model_id

    @contextmanager
    def cache_models(self):
        """Caches each trained model in memory as it is written to storage.

        Must be used as a context manager.
        The cache is cleared when the context manager goes out of scope
        """
        with self.model_storage_engine.cache_models():
            yield

    def generate_trained_models(self, grid_config, misc_db_parameters, matrix_store):
        """Train and store configured models, yielding the ids one by one

        Args:
            grid_config (dict) of format {classpath: hyperparameter dicts}
                example: { 'sklearn.ensemble.RandomForestClassifier': {
                    'n_estimators': [1,10,100,1000,10000],
                    'max_depth': [1,5,10,20,50,100],
                    'max_features': ['sqrt','log2'],
                    'min_samples_split': [2,5,10]
                } }
            misc_db_parameters (dict) params to pass through to the database
            matrix_store (catwalk.storage.MatrixStore) a matrix and metadata

        Yields: (int) model ids
        """
        for train_task in self.generate_train_tasks(
            grid_config, misc_db_parameters, matrix_store
        ):
            yield self.process_train_task(**train_task)

    def train_models(self, grid_config, misc_db_parameters, matrix_store):
        """Train and store configured models

        Args:
            grid_config (dict) of format {classpath: hyperparameter dicts}
                example: { 'sklearn.ensemble.RandomForestClassifier': {
                    'n_estimators': [1,10,100,1000,10000],
                    'max_depth': [1,5,10,20,50,100],
                    'max_features': ['sqrt','log2'],
                    'min_samples_split': [2,5,10]
                } }
            misc_db_parameters (dict) params to pass through to the database
            matrix_store(catwalk.storage.MatrixStore) a matrix and metadata

        Returns:
            (list) of model ids
        """
        return [
            model_id
            for model_id in self.generate_trained_models(
                grid_config, misc_db_parameters, matrix_store
            )
        ]

    def process_train_task(
        self, matrix_store, class_path, parameters, model_hash, misc_db_parameters, random_seed=None
    ):
        """Trains and stores a model, or skips it and returns the existing id

        Args:
            matrix_store (catwalk.storage.MatrixStore) a matrix and metadata
            class_path (string): a full class path for the classifier
            parameters (dict): all hyperparameters to be passed to the classifier
            model_hash (string) a unique id for the model
            misc_db_parameters (dict) params to pass through to the database
            random_seed (int, optional) a number to use to seed the random number generator before training. if none given, will generate one to store
        Returns: (int) model id
        """
        try:
            saved_model_id = retrieve_model_id_from_hash(self.db_engine, model_hash)
            if (
                not self.replace
                and self.model_storage_engine.exists(model_hash)
                and saved_model_id
            ):
                logger.debug(f"Skipping model {saved_model_id} {class_path} {parameters}")
                if self.run_id:
                    skipped_model(self.run_id, self.db_engine)
                return saved_model_id

            if self.replace:
                reason = "replace flag has been set"
            elif not self.model_storage_engine.exists(model_hash):
                reason = "model pickle not found in store"
            elif not saved_model_id:
                reason = "model metadata not found"

            logger.debug(
                f"Training {class_path} with parameters {parameters}"
                f"(reason to train: {reason})"
            )
            try:
                model_id = self._train_and_store_model(
                    matrix_store, class_path, parameters, model_hash, misc_db_parameters, random_seed
                )
            except BaselineFeatureNotInMatrix:
                logger.warning(
                    "Tried to train baseline model without required feature in matrix. Skipping."
                )
                if self.run_id:
                    skipped_model(self.run_id, self.db_engine)
                model_id = None
            if self.run_id:
                built_model(self.run_id, self.db_engine)
            return model_id
        except Exception as exc:
            logger.exception(f"Model training for matrix {matrix_store.uuid}, estimator {class_path}/{parameters}, model hash {model_hash} failed.")
            errored_model(self.run_id, self.db_engine)

    @staticmethod
    def flattened_grid_config(grid_config):
        return flatten_grid_config(grid_config)

    def generate_train_tasks(self, grid_config, misc_db_parameters, matrix_store=None):
        """Train and store configured models, yielding the ids one by one

        Args:
            grid_config (dict) of format {classpath: hyperparameter dicts}
                example: { 'sklearn.ensemble.RandomForestClassifier': {
                    'n_estimators': [1,10,100,1000,10000],
                    'max_depth': [1,5,10,20,50,100],
                    'max_features': ['sqrt','log2'],
                    'min_samples_split': [2,5,10]
                } }
            misc_db_parameters (dict) params to pass through to the database

        Returns: (list) training task definitions, suitable for process_train_task kwargs
        """
        matrix_store = matrix_store or self.matrix_store
        misc_db_parameters = copy.deepcopy(misc_db_parameters)
        misc_db_parameters["batch_run_time"] = datetime.datetime.now().isoformat()
        misc_db_parameters["train_end_time"] = matrix_store.metadata["end_time"]
        misc_db_parameters["training_label_timespan"] = matrix_store.metadata[
            "label_timespan"
        ]
        misc_db_parameters["train_matrix_uuid"] = matrix_store.uuid

        tasks = []

        for class_path, parameters in self.flattened_grid_config(grid_config):
            random_seed = generate_python_random_seed()
            model_hash = self._model_hash(
                matrix_store.metadata,
                class_path,
                parameters,
                random_seed
            )
            logger.spam(
                f"Computed model hash for {class_path} "
                f"with parameters {parameters}: {model_hash}"
            )

            if any(task["model_hash"] == model_hash for task in tasks):
                logger.info(
                    f"Skipping "
                    f"Classpath: {class_path}({parameters}) "
                    f"[{model_hash}] because another "
                    f"equivalent one found in this batch."
                )
                continue

            tasks.append(
                {
                    "matrix_store": matrix_store,
                    "class_path": class_path,
                    "parameters": parameters,
                    "model_hash": model_hash,
                    "misc_db_parameters": misc_db_parameters,
                    "random_seed": random_seed
                }
            )
            logger.debug(f"Task added for model {class_path}({parameters}) [{model_hash}]")
        logger.debug(f"Found {len(tasks)} unique model training tasks")
        return tasks



import verboselogs, logging
logger = verboselogs.VerboseLogger(__name__)

from collections import OrderedDict

import sqlalchemy
import sqlparse

from triage.util.conf import convert_str_to_relativedelta
from triage.database_reflection import table_exists

from triage.component.collate import (
    Aggregate,
    Categorical,
    Compare,
    SpacetimeAggregation,
    FromObj
)

class FeatureGenerator:
    def __init__(
        self,
        db_engine,
        features_schema_name,
        replace=True,
        feature_start_time=None,
        materialize_subquery_fromobjs=True,
        features_ignore_cohort=False,
    ):
        """Generates aggregate features using collate

        Args:
            db_engine (sqlalchemy.db.engine)
            features_schema_name (string) Name of schema where feature
                tables should be written to
            replace (boolean, optional) Whether or not existing features
                should be replaced
            feature_start_time (string/datetime, optional) point in time before which
                should not be included in features
            features_ignore_cohort (boolean, optional) Whether or not features should be built
                independently of the cohort. Takes longer but means that features can be reused
                for different cohorts.
        """
        self.db_engine = db_engine
        self.features_schema_name = features_schema_name
        self.categorical_cache = {}
        self.replace = replace
        self.feature_start_time = feature_start_time
        self.materialize_subquery_fromobjs = materialize_subquery_fromobjs
        self.features_ignore_cohort = features_ignore_cohort
        self.entity_id_column = "entity_id"
        self.from_objs = {}

    def _validate_keys(self, aggregation_config):
        for key in [
            "from_obj",
            "intervals",
            "groups",
            "knowledge_date_column",
            "prefix",
        ]:
            if key not in aggregation_config:
                raise ValueError(
                    "{} required as key: aggregation config: {}".format(
                        key, aggregation_config
                    )
                )

    def _validate_aggregates(self, aggregation_config):
        if (
            "aggregates" not in aggregation_config
            and "categoricals" not in aggregation_config
            and "array_categoricals" not in aggregation_config
        ):
            raise ValueError(
                "Need either aggregates, categoricals, or array_categoricals"
                + " in {}".format(aggregation_config)
            )

    def _validate_categoricals(self, categoricals):
        for categorical in categoricals:
            if "choice_query" in categorical:
                logger.spam("Validating choice query")

                try:
                    with self.db_engine.begin() as conn:
                        conn.execute("explain {}".format(categorical["choice_query"]))
                except Exception as exc:
                    raise ValueError(
                        f"choice query does not run. \n"
                        f'choice query: {categorical["choice_query"]}\n'
                    )

    def _validate_from_obj(self, from_obj):
        logger.spam("Validating from_obj")
        try:
            with self.db_engine.begin() as conn:
                conn.execute("explain select * from {}".format(from_obj))
        except Exception as exc:
            raise ValueError(
                "from_obj query does not run. \n"
                'from_obj: "{from_obj}"\n'
            )

    def _validate_time_intervals(self, intervals):
        logger.spam("Validating time intervals")
        for interval in intervals:
            if interval != "all":
                convert_str_to_relativedelta(interval)

    def _validate_groups(self, groups):
        logger.spam("Validating groups")
        if "entity_id" not in groups:
            raise ValueError(
                "One of the aggregation groups is required to be entity_id"
            )

    def _validate_imputation_rule(self, aggregate_type, impute_rule):
        """Validate the imputation rule for a given aggregation type."""
        logger.spam("Validating imputation rule")
        # dictionary of imputation type : required parameters
        valid_imputations = {
            "all": {
                "mean": [],
                "constant": ["value"],
                "zero": [],
                "zero_noflag": [],
                "error": [],
            },
            "aggregates": {"binary_mode": []},
            "categoricals": {"null_category": []},
        }
        valid_imputations["array_categoricals"] = valid_imputations["categoricals"]

        # the valid imputation rules for the specific aggregation type being checked
        valid_types = dict(
            valid_imputations["all"], **valid_imputations[aggregate_type]
        )

        # no imputation rule was specified
        if "type" not in impute_rule.keys():
            raise ValueError("Imputation type must be specified")

        # a rule was specified, but not valid for this type of aggregate
        if impute_rule["type"] not in valid_types.keys():
            raise ValueError(
                f"Invalid imputation type {impute_rule['type']} for {aggregate_type}"
            )

        # check that all required parameters exist in the keys of the imputation rule
        required_params = valid_types[impute_rule["type"]]
        for param in required_params:
            if param not in impute_rule.keys():
                raise ValueError(
                    "Missing param {param} for {imput_rule['type']}"
                )

    def _validate_imputations(self, aggregation_config):
        """Validate the imputation rules in an aggregation config, looping
        through all three types of aggregates. Most of the work here is
        done by _validate_imputation_rule() to check the requirements of
        each imputation rule found
        """
        logger.spam("Validating imputations")
        agg_types = ["aggregates", "categoricals", "array_categoricals"]

        for agg_type in agg_types:
            # base_imp are the top-level rules, `such as aggregates_imputation`
            base_imp = aggregation_config.get(agg_type + "_imputation", {})

            # loop through the individual aggregates
            for agg in aggregation_config.get(agg_type, []):
                # combine any aggregate-level imputation rules with top-level ones
                imp_dict = dict(base_imp, **agg.get("imputation", {}))

                # imputation rules are metric-specific, so check each metric's rule
                for metric in agg["metrics"]:
                    # metric rules may be defined by the metric name (e.g., 'max')
                    # or with the 'all' catch-all, with named metrics taking
                    # precedence. If we fall back to {}, the rule validator will
                    # error out on no metric found.
                    impute_rule = imp_dict.get(metric, imp_dict.get("all", {}))
                    self._validate_imputation_rule(agg_type, impute_rule)

    def _validate_aggregation(self, aggregation_config):
        logger.spam(f"Validating aggregation config {aggregation_config}")
        self._validate_keys(aggregation_config)
        self._validate_aggregates(aggregation_config)
        self._validate_categoricals(aggregation_config.get("categoricals", []))
        self._validate_from_obj(aggregation_config["from_obj"])
        self._validate_time_intervals(aggregation_config["intervals"])
        self._validate_groups(aggregation_config["groups"])
        self._validate_imputations(aggregation_config)

    def validate(self, feature_aggregation_config):
        """Validate a feature aggregation config applied to this object

        The validations range from basic type checks, key presence checks,
        as well as validating the sql in from objects.

        Args:
            feature_aggregation_config (list) all values, except for feature
                date, necessary to instantiate a collate.SpacetimeAggregation

        Raises: ValueError if any part of the config is found to be invalid
        """
        logger.spam(f"Validating feature aggregation configurations")
        for aggregation in feature_aggregation_config:
            self._validate_aggregation(aggregation)

    def _compute_choices(self, choice_query):
        if choice_query not in self.categorical_cache:
            with self.db_engine.begin() as conn:
                self.categorical_cache[choice_query] = [
                    row[0] for row in conn.execute(choice_query)
                ]

            logger.debug(
                "Computed list of categoricals: {self.categorical_cache[choice_query]} for choice query: {choice_query}"
            )

        return self.categorical_cache[choice_query]

    def _build_choices(self, categorical):
        logger.debug(
            f'Building categorical choices for column {categorical["column"]}, metrics {categorical["metrics"]}',
        )
        if "choices" in categorical:
            logger.debug(f'Found list of configured choices: {categorical["choices"]}')
            return categorical["choices"]
        else:
            return self._compute_choices(categorical["choice_query"])

    def _build_categoricals(self, categorical_config, impute_rules):
        # TODO: only include null flag where necessary
        return [
            Categorical(
                col=categorical["column"],
                choices=self._build_choices(categorical),
                function=categorical["metrics"],
                impute_rules=dict(
                    impute_rules,
                    coltype="categorical",
                    **categorical.get("imputation", {})
                ),
                include_null=True,
                coltype=categorical.get('coltype', None),
            )
            for categorical in categorical_config
        ]

    def _build_array_categoricals(self, categorical_config, impute_rules):
        # TODO: only include null flag where necessary
        return [
            Compare(
                col=categorical["column"],
                op="@>",
                choices={
                    choice: "array['{}'::varchar]".format(choice)
                    for choice in self._build_choices(categorical)
                },
                function=categorical["metrics"],
                impute_rules=dict(
                    impute_rules,
                    coltype="array_categorical",
                    **categorical.get("imputation", {})
                ),
                op_in_name=False,
                quote_choices=False,
                include_null=True,
                coltype=categorical.get('coltype', None)
            )
            for categorical in categorical_config
        ]

    def _aggregation(self, aggregation_config, feature_dates, state_table):
        logger.debug(
            f"Building collate.SpacetimeAggregation for config {aggregation_config} and {len(feature_dates)} as_of_dates",
        )

        # read top-level imputation rules from the aggregation config; we'll allow
        # these to be overridden by imputation rules at the individual feature
        # level as those get parsed as well
        agimp = aggregation_config.get("aggregates_imputation", {})
        catimp = aggregation_config.get("categoricals_imputation", {})
        arrcatimp = aggregation_config.get("array_categoricals_imputation", {})

        aggregates = [
            Aggregate(
                aggregate["quantity"],
                aggregate["metrics"],
                dict(agimp, coltype="aggregate", **aggregate.get("imputation", {})),
                coltype=aggregate.get('coltype', None)
            )
            for aggregate in aggregation_config.get("aggregates", [])
        ]
        logger.debug(f"Found {len(aggregates)} quantity aggregates")
        categoricals = self._build_categoricals(
            aggregation_config.get("categoricals", []), catimp
        )
        logger.debug(f"Found {len(categoricals)} categorical aggregates")
        array_categoricals = self._build_array_categoricals(
            aggregation_config.get("array_categoricals", []), arrcatimp
        )
        logger.debug(f"Found {len(array_categoricals)} array categorical aggregates")
        return SpacetimeAggregation(
            aggregates + categoricals + array_categoricals,
            from_obj=aggregation_config["from_obj"],
            intervals=aggregation_config["intervals"],
            groups=aggregation_config["groups"],
            dates=feature_dates,
            state_table=state_table,
            state_group=self.entity_id_column,
            date_column=aggregation_config["knowledge_date_column"],
            output_date_column="as_of_date",
            input_min_date=self.feature_start_time,
            schema=self.features_schema_name,
            prefix=aggregation_config["prefix"],
            join_with_cohort_table=not self.features_ignore_cohort
        )

    def aggregations(self, feature_aggregation_config, feature_dates, state_table):
        """Creates collate.SpacetimeAggregations from the given arguments

        Args:
            feature_aggregation_config (list) all values, except for feature
                date, necessary to instantiate a collate.SpacetimeAggregation
            feature_dates (list) dates to generate features as of
            state_table (string) schema.table_name for state table with all entity/date pairs

        Returns: (list) collate.SpacetimeAggregations
        """
        return [
            self.preprocess_aggregation(
                self._aggregation(aggregation_config, feature_dates, state_table)
            )
            for aggregation_config in feature_aggregation_config
        ]

    def preprocess_aggregation(self, aggregation):
        create_schema = aggregation.get_create_schema()

        if create_schema is not None:
            with self.db_engine.begin() as conn:
                conn.execute(create_schema)

        if self.materialize_subquery_fromobjs:
            # materialize from obj
            from_obj = FromObj(
                from_obj=aggregation.from_obj.text,
                name=f"{aggregation.schema}.{aggregation.prefix}",
                knowledge_date_column=aggregation.date_column
            )
            from_obj.maybe_materialize(self.db_engine)
            aggregation.from_obj = from_obj.table
        return aggregation

    def generate_all_table_tasks(self, aggregations, task_type):
        """Generates SQL commands for creating, populating, and indexing
        feature group tables

        Args:
            aggregations (list) collate.SpacetimeAggregation objects
            type (str) either 'aggregation' or 'imputation'

        Returns: (dict) keys are group table names, values are themselves dicts,
            each with keys for different stages of table creation (prepare, inserts, finalize)
            and with values being lists of SQL commands
        """

        # pick the method to use for generating tasks depending on whether we're
        # building the aggregations or imputations
        if task_type == "aggregation":
            task_generator = self._generate_agg_table_tasks_for
            logger.verbose("Starting Feature aggregation")
        elif task_type == "imputation":
            task_generator = self._generate_imp_table_tasks_for
            logger.verbose("Starting Feature imputation")
        else:
            raise ValueError(f"Table task type must be aggregation or imputation, was: {table_task}")

        table_tasks = OrderedDict()
        for aggregation in aggregations:
            table_tasks.update(task_generator(aggregation))
        return table_tasks

    def create_features_before_imputation(
        self, feature_aggregation_config, feature_dates, state_table=None
    ):
        """Create features before imputation for a set of dates"""
        all_tasks = self.generate_all_table_tasks(
            self.aggregations(
                feature_aggregation_config, feature_dates, state_table=state_table
            ),
            task_type="aggregation",
        )
        logger.debug(f"Generated a total of {len(all_tasks)} table tasks")
        for task_num, task in enumerate(all_tasks.values(), 1):
            prepares = task.get("prepare", [])
            inserts = task.get("inserts", [])
            finalize = task.get("finalize", [])
            logger.verbose(f"Executing task [{task_num} / {len(all_tasks)}]")
            logger.verbose(
                f"{len(prepares)} prepare queries, {len(inserts)} insert queries, {len(finalize)} finalize queries"
            )
            logger.verbose("Executing {len(prepares)} preparation queries ")
            for query_num, query in enumerate(prepares, 1):
                logger.verbose(f"Prepare query {query_num} / {len(prepares)}")
                logger.debug(
                    f"Prepare query {query_num}: {sqlparse.format(str(query), reindent=True)}",
                )
            logger.verbose("Executing {len(inserts)} insert queries ")
            for query_num, query in enumerate(inserts, 1):
                logger.verbose(f"Insert query {query_num} / {len(inserts)}")
                logger.debug(
                    f"Insert query {query_num}: {sqlparse.format(str(query), reindent=True)}"
                )
            logger.verbose("Executing {len(finalize)} finalize queries ")
            for query_num, query in enumerate(finalize, 1):
                logger.verbose(f"Finalize query {query_num} / {len(finalize)}")
                logger.debug(
                    f"Finalize query {query_num}: {sqlparse.format(str(query), reindent=True)}"
                )
            self.process_table_task(task)

    def create_all_tables(self, feature_aggregation_config, feature_dates, state_table):
        """Create all feature tables.

        First builds the aggregation tables, and then performs
        imputation on any null values, (requiring a two-step process to
        determine which columns contain nulls after the initial
        aggregation tables are built).

        Args:
            feature_aggregation_config (list) all values, except for
                feature date, necessary to instantiate a
                `collate.SpacetimeAggregation`
            feature_dates (list) dates to generate features as of
            state_table (string) schema.table_name for state table with
                all entity/date pairs

        Returns: (list) table names

        """
        aggs = self.aggregations(feature_aggregation_config, feature_dates, state_table)

        # first, generate and run table tasks for aggregations
        table_tasks_aggregate = self.generate_all_table_tasks(aggs, task_type="aggregation")
        self.process_table_tasks(table_tasks_aggregate)

        # second, perform the imputations (this will query the tables
        # constructed above to identify features containing nulls)
        table_tasks_impute = self.generate_all_table_tasks(aggs, task_type="imputation")
        impute_keys = self.process_table_tasks(table_tasks_impute)

        # double-check that the imputation worked and no nulls remain
        # in the data:
        nullcols = []
        with self.db_engine.begin() as conn:
            for agg in aggs:
                results = conn.execute(agg.find_nulls(imputed=True))
                null_counts = results.first().items()
                nullcols += [col for (col, val) in null_counts if val > 0]

        if len(nullcols) > 0:
            raise ValueError(
                f"Imputation failed for {len(nullcols)} columns. Null values remain in: {nullcols}"
            )

        return impute_keys

    def process_table_task(self, task):
        self.run_commands(task.get("prepare", []))
        self.run_commands(task.get("inserts", []))
        self.run_commands(task.get("finalize", []))

    def process_table_tasks(self, table_tasks):
        for table_name, task in table_tasks.items():
            self.process_table_task(task)
        return table_tasks.keys()

    def _explain_selects(self, aggregations):
        with self.db_engine.begin() as conn:
            for aggregation in aggregations:
                for selectlist in aggregation.get_selects().values():
                    for select in selectlist:
                        query = "explain " + str(select)
                        results = list(conn.execute(query))
                        logger.spam(str(select))
                        logger.spam(results)

    def _clean_table_name(self, table_name):
        # remove the schema and quotes from the name
        return table_name.split(".")[1].replace('"', "")

    def _table_exists(self, table_name):
        try:
            with self.db_engine.begin() as conn:
                conn.execute(
                    "select 1 from {}.{} limit 1".format(
                        self.features_schema_name, table_name
                    )
                ).first()
        except sqlalchemy.exc.ProgrammingError:
            return False
        else:
            return True

    def run_commands(self, command_list):
        with self.db_engine.begin() as conn:
            for command in command_list:
                logger.spam(f"Executing feature generation query: {command}")
                conn.execute(command)

    def _aggregation_index_query(self, aggregation, imputed=False):
        return f"CREATE INDEX ON {aggregation.get_table_name(imputed=imputed)} ({self.entity_id_column}, {aggregation.output_date_column})"

    def _aggregation_index_columns(self, aggregation):
        return sorted(
            [group for group in aggregation.groups.keys()]
            + [aggregation.output_date_column]
        )

    def index_column_lookup(self, aggregations, imputed=True):
        return dict(
            (
                self._clean_table_name(aggregation.get_table_name(imputed=imputed)),
                self._aggregation_index_columns(aggregation),
            )
            for aggregation in aggregations
        )

    def _needs_features(self, aggregation):
        imputed_table = self._clean_table_name(
            aggregation.get_table_name(imputed=True)
        )

        if self._table_exists(imputed_table):
            check_query = (
                f"select 1 from {aggregation.state_table} "
                f"left join {self.features_schema_name}.{imputed_table} "
                f"using (entity_id, as_of_date) "
                f"where {self.features_schema_name}.{imputed_table}.entity_id is null limit 1"
            )
            if self.db_engine.execute(check_query).scalar():
                logger.notice(
                    f"Imputed feature table {imputed_table} did not contain rows from the "
                    f"entire cohort, need to rebuild features"
                )
                return True
        else:
            logger.notice(
                f"Imputed feature table {imputed_table} did not exist, "
                f"need to build features"
            )
            return True
        logger.notice(f"Imputed feature table {imputed_table} looks good, "
                      f"skipping feature building!")
        return False

    def _generate_agg_table_tasks_for(self, aggregation):
        """Generates SQL commands for preparing, populating, and finalizing
        each feature group table in the given aggregation

        Args:
            aggregation (collate.SpacetimeAggregation)

        Returns: (dict) of structure {
            'prepare': list of commands to prepare table for population
            'inserts': list of commands to populate table
            'finalize': list of commands to finalize table after population
        }
        """
        creates = aggregation.get_creates()
        drops = aggregation.get_drops()
        indexes = aggregation.get_indexes()
        inserts = aggregation.get_inserts()
        table_tasks = OrderedDict()

        needs_features = self._needs_features(aggregation)

        for group in aggregation.groups:
            group_table = self._clean_table_name(
                aggregation.get_table_name(group=group)
            )
            if self.replace or needs_features:
                table_tasks[group_table] = {
                    "prepare": [drops[group], creates[group]],
                    "inserts": inserts[group],
                    "finalize": [indexes[group]],
                }
                logger.debug(f"Created task for table {group_table}")
            else:
                logger.debug(f"Skipping feature creation for table {group_table}")
                table_tasks[group_table] = {}
        if self.replace or needs_features:
            table_tasks[self._clean_table_name(aggregation.get_table_name())] = {
                "prepare": [aggregation.get_drop(), aggregation.get_create()],
                "inserts": [],
                "finalize": [self._aggregation_index_query(aggregation)],
            }
            logger.debug(f"Created tasks for aggregation {self._clean_table_name(aggregation.get_table_name())}" )
        else:
            logger.debug(f"Skipping feature creation for table {self._clean_table_name(aggregation.get_table_name())}")
            table_tasks[self._clean_table_name(aggregation.get_table_name())] = {}

        return table_tasks

    def _generate_imp_table_tasks_for(self, aggregation, drop_preagg=True):
        """Generate SQL statements for preparing, populating, and
        finalizing imputations, for each feature group table in the
        given aggregation.

        Requires the existance of the underlying feature and aggregation
        tables defined in `_generate_agg_table_tasks_for()`.

        Args:
            aggregation (collate.SpacetimeAggregation)
            drop_preagg: boolean to specify dropping pre-imputation
                tables

        Returns: (dict) of structure {
                'prepare': list of commands to prepare table for population
                'inserts': list of commands to populate table
                'finalize': list of commands to finalize table after population
            }

        """
        table_tasks = OrderedDict()
        imp_tbl_name = self._clean_table_name(aggregation.get_table_name(imputed=True))

        if not self.replace and not self._needs_features(aggregation):
            logger.debug("Skipping imputation table creation for %s", imp_tbl_name)
            table_tasks[imp_tbl_name] = {}
            return table_tasks

        if not aggregation.state_table:
            logger.debug(
                f"No state table defined in aggregation, cannot create imputation table for {imp_tbl_name}",
            )
            table_tasks[imp_tbl_name] = {}
            return table_tasks

        if not table_exists(aggregation.state_table, self.db_engine):
            logger.debug(
                f"State table {aggregation.state_table} does not exist, cannot create imputation table for {imp_tbl_name}",
            )
            table_tasks[imp_tbl_name] = {}
            return table_tasks

        # excute query to find columns with null values and create lists of columns
        # that do and do not need imputation when creating the imputation table
        with self.db_engine.begin() as conn:
            results = conn.execute(aggregation.find_nulls())
            null_counts = results.first().items()
        impute_cols = [col for (col, val) in null_counts if val > 0]
        nonimpute_cols = [col for (col, val) in null_counts if val == 0]

        # table tasks for imputed aggregation table, most of the work is done here
        # by collate's get_impute_create()
        table_tasks[imp_tbl_name] = {
            "prepare": [
                aggregation.get_drop(imputed=True),
                aggregation.get_impute_create(
                    impute_cols=impute_cols, nonimpute_cols=nonimpute_cols
                ),
            ],
            "inserts": [],
            "finalize": [self._aggregation_index_query(aggregation, imputed=True)],
        }
        logger.debug("Created table tasks for imputation: %s", imp_tbl_name)

        # do some cleanup:
        # drop the group-level and aggregation tables, just leaving the
        # imputation table if drop_preagg=True
        if drop_preagg:
            drops = aggregation.get_drops()
            table_tasks[imp_tbl_name]["finalize"] += list(drops.values()) + [
                aggregation.get_drop()
            ]
            logger.debug("Added drop table cleanup tasks: %s", imp_tbl_name)

        return table_tasks



"""
Model Evaluator

This script contain a set of elements to help the postmodeling evaluation
of audited models by triage.Audition. This will be a continuing list of
routines that can be scaled up and grow according to the needs of the
project, or other postmodeling approaches.

"""
import itertools
import pandas as pd
import numpy as np
import seaborn as sns
from descriptors import cachedproperty
from matplotlib import pyplot as plt
from itertools import combinations
from scipy.spatial.distance import squareform, pdist
from scipy.stats import spearmanr

# Get indivual model information/metadata from Audition output


class ModelGroupEvaluator:
    '''
    ModelGroup class calls the model group metadata from the database
    and hold metadata features on each of the class attibutes.
    This class will contain any information about the model_group, and
    will be used to make comparisons and calculations across models.

    A model_group_id list is needed to instate the class.
    '''
    def __init__(self, model_group_id, engine):

        self.engine = engine

        if len(model_group_id) == 1:
            self.model_group_id = model_group_id + model_group_id
        else:
            self.model_group_id = model_group_id

    @cachedproperty
    def metadata(self):
        query_execute = list(self.engine.execute(
            f'''
            SELECT m.model_id,
                   m.model_group_id,
                   m.hyperparameters,
                   m.model_hash,
                   m.train_end_time,
                   m.train_matrix_uuid,
                   m.training_label_timespan,
                   m.model_type,
                   mg.model_config
                FROM triage_metadata.models m
                JOIN triage_metadata.model_groups mg
                USING (model_group_id)
                WHERE model_group_id IN {self.model_group_id}
            ''')
        )

        row_dict, list_dict = {}, []
        for row in query_execute:
            for tup in row.items():
                row_dict = {**row_dict, **{tup[0]: tup[1]}}
            list_dict.append(row_dict)

        return list_dict

    @property
    def model_id(self):
        return [dict_row['model_id'] for dict_row in self.metadata]

    @property
    def model_hash(self):
        return [dict_row['model_hash'] for dict_row in self.metadata]

    @property
    def hyperparameters(self):
        return [dict_row['hyperparameters'] for dict_row in self.metadata]

    @property
    def train_end_time(self):
        return [dict_row['train_end_time'] for dict_row in self.metadata]

    @property
    def train_matrix_uuid(self):
        return [dict_row['train_matrix_uuid'] for dict_row in self.metadata]

    @property
    def training_label_timespan(self):
        return [dict_row['training_label_timespan'] for dict_row in self.metadata]

    @property
    def model_type(self):
        return [dict_row['model_type'] for dict_row in self.metadata]

    @property
    def model_config(self):
        return [dict_row['model_config'] for dict_row in self.metadata]

    def __repr__(self):
        return f"ModelGroupEvaluator(model_group_id={self.model_group_id})"

    @cachedproperty
    def predictions(self):
        preds = pd.read_sql(
            f'''
            SELECT
                   g.model_group_id,
                   m.model_id,
                   m.entity_id,
                   m.as_of_date,
                   EXTRACT('YEAR' from m.as_of_date) AS as_of_date_year,
                   m.score,
                   m.label_value,
                   COALESCE(rank_abs_with_ties, RANK() OVER(PARTITION BY m.model_id
                   ORDER BY m.score DESC)) AS rank_abs,
                   COALESCE(m.rank_pct_with_ties, percent_rank() OVER(PARTITION BY
                   m.model_id ORDER BY m.score DESC)) * 100 AS rank_pct,
                   m.test_label_timespan
            FROM test_results.predictions m
            LEFT JOIN triage_metadata.models g
            USING (model_id)
            WHERE model_id IN {tuple(self.model_id)}
            AND label_value IS NOT NULL
            ''', con=self.engine)
        if preds.empty:
            raise RuntimeError("No predictions were retrieved from the database."
                               "Some functionality will not be available without predictions."
                               "Please run catwalk.Predictor for each desired model and test matrix"
                               )
        return preds

    @cachedproperty
    def feature_importances(self):
        features = pd.read_sql(
           f'''
           SELECT g.model_group_id,
                  m.model_id,
                  m.feature,
                  m.feature_importance,
                  m.rank_abs
           FROM train_results.feature_importances m
           LEFT JOIN triage_metadata.models g
           USING (model_id)
           WHERE m.model_id IN {tuple(self.model_id)}
           ''', con=self.engine)
        return features

    @cachedproperty
    def metrics(self):
        model_metrics = pd.read_sql(
            f'''
            SELECT g.model_group_id,
                   m.model_id,
                   EXTRACT('YEAR' FROM m.evaluation_end_time) AS as_of_date_year,
                   m.metric,
                   m.parameter,
                   m.stochastic_value as value,
                   m.num_labeled_examples,
                   m.num_labeled_above_threshold,
                   m.num_positive_labels
            FROM test_results.evaluations m
            LEFT JOIN triage_metadata.models g
            USING (model_id)
            WHERE m.model_id IN {tuple(self.model_id)}
            ''', con=self.engine)
        return model_metrics

    @cachedproperty
    def feature_groups(self):
        model_feature_groups = pd.read_sql(
            f'''
            WITH
            feature_groups_raw AS(
            SELECT
            model_group_id,
            model_config->>'feature_groups' as features
            FROM triage_metadata.model_groups
            WHERE model_group_id IN {self.model_group_id}
            ),
            feature_groups_unnest AS(
            SELECT model_group_id,
            unnest(regexp_split_to_array(substring(features, '\[(.*?)\]'), ',')) AS group_array
            FROM feature_groups_raw
            ), feature_groups_array AS(
            SELECT
            model_group_id,
            array_agg(split_part(substring(group_array,
            '\"(.*?)\"'), ':', 2)) AS feature_group_array
            FROM feature_groups_unnest
            GROUP BY model_group_id
            ), feature_groups_array_ AS(
            SELECT
            model_group_id,
            feature_group_array,
            array_length(feature_group_array, 1) AS number_feature_groups
            FROM feature_groups_array
            ), feature_groups_class_cases
            AS(
            SELECT
            model_group_id,
            feature_group_array,
            number_feature_groups,
            CASE
            WHEN number_feature_groups = 1
            THEN 'LOI'
            WHEN  number_feature_groups = first_value(number_feature_groups) OVER w
            THEN 'All features'
            WHEN  number_feature_groups = (first_value(number_feature_groups) OVER w) - 1
            THEN 'LOO'
            ELSE NULL
            END AS experiment_type
            FROM feature_groups_array_
            WINDOW w AS (ORDER BY number_feature_groups DESC)
            ) SELECT * FROM feature_groups_class_cases
            ''', con=self.engine)
        return model_feature_groups

    @cachedproperty
    def same_time_models(self):
        time_models = pd.read_sql(
                      f'''
                      SELECT
                      train_end_time,
                      array_agg(model_id) AS model_id_array
                      FROM triage_metadata.models
                      WHERE model_group_id IN {self.model_group_id}
                      GROUP BY train_end_time
                      ''', con=self.engine)
        return time_models

    def plot_prec_across_time(self,
                              param_type=None,
                              param=None,
                              metric=None,
                              baseline=False,
                              baseline_query=None,
                              df=False,
                              figsize=(12, 16),
                              fontsize=20):

        '''
        Plot precision across time for all model_group_ids, and baseline,
        if available.

        This function plots the performance of each model_group_id following an
        user defined performance metric. First, this function check if the
        performance metrics are available to both the models, and the baseline.
        Second, filter the data of interest, and lastly, plot the results as
        timelines (model_id date).

        Arguments:
            - param_type (string): A parameter string with a threshold
            definition: rank_pct, or rank_abs. These are usually defined in the
            postmodeling configuration file.
            - param (int): A threshold value compatible with the param_type.
            This value is also defined in the postmodeling configuration file.
            - metric (string): A string defnining the type of metric use to
            evaluate the model, this can be 'precision@', or 'recall@'.
            - baseline (bool): should we include a baseline for comparison?
            - baseline_query (str): a SQL query that returns the evaluation data from
            the baseline models. This value can also be defined in the
            configuration file.
            - df (bool): If True, no plot is rendered, but a pandas DataFrame
            is returned with the data.
            - figsize (tuple): tuple with figure size parameters
            - fontsize (int): Fontsize for titles
        '''

        # Load metrics and prepare data for analysis
        model_metrics = self.metrics
        model_metrics[['param', 'param_type']] = \
                model_metrics['parameter'].str.split('_', 1, expand=True)
        model_metrics['param'] =  model_metrics['param'].replace('',np.nan).astype('float')
        model_metrics['param_type'] = model_metrics['param_type'].apply(lambda x: f'rank_{x}' if x is not None else x)

        # Filter model_group_id metrics and create pivot table by each
        # model_group_id.
        model_metrics_filter = model_metrics[(model_metrics['metric'] == metric) &
                                      (model_metrics['param'] == param) &
                                      (model_metrics['param_type'] == param_type)].\
                filter(['model_group_id', 'model_id', 'as_of_date_year',
                        'value'])

        if baseline == True:

            baseline_metrics = pd.read_sql(baseline_query, con=self.engine)
            baseline_metrics[['param', 'param_type']] = \
                    baseline_metrics['parameter'].str.split('_', 1, expand=True)
            baseline_metrics['param'] = baseline_metrics['param'].astype(str).astype(float)
            baseline_metrics['param_type'] = baseline_metrics['param_type'].\
                    apply(lambda x: 'rank_'+x)

            # Filter baseline metrics and create pivot table to join with
            # selected models metrics
            baseline_metrics_filter =  baseline_metrics.\
                    filter(['model_group_id', 'model_id', 'as_of_date_year', 'value'])

            baseline_metrics_filter['model_group_id'] = \
                    baseline_metrics_filter.model_group_id.\
                    apply(lambda x: 'baseline_' + str(x))

            # Join tables by index(as_of_date_year)
            model_metrics_filter = \
            model_metrics_filter.append(baseline_metrics_filter, sort=True)

        model_metrics_filter['as_of_date_year'] = \
                model_metrics_filter.as_of_date_year.astype('int')
        model_metrics_filter = model_metrics_filter.sort_values('as_of_date_year')

        if df == True:
            return model_metrics_filter

        else:
            try:
                sns.set_style('whitegrid')
                fig, ax = plt.subplots(figsize=figsize)
                for model_group, df in model_metrics_filter.groupby(['model_group_id']):
                    ax = ax = df.plot(ax=ax, kind='line',
                                      x='as_of_date_year',
                                      y='value',
                                      marker='d',
                                      label=model_group)
                plt.title(str(metric).capitalize() +\
                          ' for selected model_groups in time.',
                          fontsize=fontsize)
                ax.tick_params(labelsize=16)
                ax.set_xlabel('Year of prediction (as_of_date)', fontsize=20)
                ax.set_ylabel(f'{str(metric)+str(param_type)+str(param)}',
                              fontsize=20)
                plt.xticks(model_metrics_filter.as_of_date_year.unique())
                plt.yticks(np.arange(0,1,0.1))
                legend=plt.legend(bbox_to_anchor=(1.05, 1),
                           loc=2,
                           borderaxespad=0.,
                           title='Model Group',
                           fontsize=fontsize)
                legend.get_title().set_fontsize('16')

            except TypeError:
                    print(f'''
                          Oops! model_metrics_pivot table is empty. Several problems
                          can be creating this error:
                          1. Check that {param_type}@{param} exists in the evaluations
                          table
                          2. Check that the metric {metric} is available to the
                          specified {param_type}@{param}.
                          3. You basline model can have different specifications.
                          Check those!
                          4. Check overlap between baseline dates and model dates.
                          The join is using the dates for doing these, and it's
                          possible that your timestamps differ.
                          ''')

    def feature_loi_loo(self,
                        model_subset=None,
                        param_type=None,
                        param=None,
                        metric=None,
                        baseline=False,
                        baseline_query=None,
                        df=False,
                        figsize=(16,12),
                        fontsize=20):
        '''
        Plot precision across time for the selected model_group_ids, and
        include a leave-one-out, leave-one-in feature analysis to explore the
        leverage/changes of the selected metric across different models.

        This function plots the performance of each model_group_id following an
        user defined performance metric. First, this function check if the
        performance metrics are available to both the models, and the baseline.
        Second, filter the data of interest, and lastly, plot the results as
        timelines (model_id date).

        Arguments:
            -model_subset (list): A list of model_group_ids, in case you want
            to override the selected models.
            - param_type (string): A parameter string with a threshold
            definition: rank_pct, or rank_abs. These are usually defined in the
            postmodeling configuration file.
            - param (int): A threshold value compatible with the param_type.
            This value is also defined in the postmodeling configuration file.
            - metric (string): A string defnining the type of metric use to
            evaluate the model, this can be 'precision@', or 'recall@'.
            - baseline (bool): should we include a baseline for comparison?
            - baseline_query (str): a SQL query that returns the evaluation data from
            the baseline models. This value can also be defined in the
            configuration file.
            - df (bool): If True, no plot is rendered, but a pandas DataFrame
            is returned with the data.
            - figsize (tuple): tuple with figure size parameters
            - fontsize (int): Fontsize for titles

    '''

        if model_subset is None:
            model_subset = self.model_group_id

        # Load feature groups and subset
        feature_groups = self.feature_groups
        feature_groups_filter = \
        feature_groups[feature_groups['model_group_id'].isin(model_subset)]

        # Format metrics columns and filter by metric of interest
        model_metrics = self.metrics
        model_metrics[['param', 'param_type']] = \
                model_metrics['parameter'].str.split('_', 1, expand=True)
        model_metrics['param'] =  model_metrics['param'].astype(str).astype(float)
        model_metrics['param_type'] = model_metrics['param_type'].apply(lambda x: 'rank_'+x)

        model_metrics_filter = model_metrics[(model_metrics['metric'] == metric) &
                                             (model_metrics['param'] == param) &
                                             (model_metrics['param_type'] == param_type)].\
                filter(['model_group_id', 'model_id', 'as_of_date_year',
                        'value'])

        # Merge metrics and features and filter by threshold definition
        metrics_merge = model_metrics_filter.merge(feature_groups_filter,
                                                  how='inner',
                                                  on='model_group_id')

        # LOO and LOI definition
        all_features = set(metrics_merge.loc[metrics_merge['experiment_type'] == 'All features'] \
                           ['feature_group_array'][0])

        metrics_merge_experimental = \
        metrics_merge[metrics_merge['experiment_type'] != 'All features']

        metrics_merge_experimental['feature_experiment'] = \
                metrics_merge_experimental.apply(lambda row: \
                                   list(all_features.difference(row['feature_group_array']))[0] + '_loo' \
                                   if row['experiment_type'] == 'LOO' \
                                   else row['feature_group_array'] + '_loi', axis=1)

        metrics_merge_experimental = \
        metrics_merge_experimental.filter(['feature_experiment',
                                           'as_of_date_year', 'value'])

        if baseline == True:

            baseline_metrics = pd.read_sql(baseline_query, con=self.engine)
            baseline_metrics[['param', 'param_type']] = \
                    baseline_metrics['parameter'].str.split('_', 1, expand=True)
            baseline_metrics['param'] = baseline_metrics['param'].astype(str).astype(float)
            baseline_metrics['param_type'] = baseline_metrics['param_type'].apply(lambda x: 'rank_'+x)

            # Filter baseline metrics and create pivot table to join with
            # selected models metrics
            baseline_metrics_filter =  baseline_metrics[(baseline_metrics['metric'] == metric) &
                                                        (baseline_metrics['param'] == param) &
                                                        (baseline_metrics['param_type'] == param_type)].\
                filter(['model_group_id', 'model_id', 'as_of_date_year', 'value'])

            baseline_metrics_filter['feature_experiment'] = \
                    baseline_metrics_filter.model_group_id.apply(lambda x: \
                                                                 'baseline_' + \
                                                                 str(x))

            metrics_merge_experimental = \
                    metrics_merge_experimental.append(baseline_metrics_filter,
                                                     sort=True)
        if df == True:
            return metrics_merge_experimental

        else:
            try:
                sns.set_style('whitegrid')
                fig, ax = plt.subplots(figsize=figsize)
                for feature, df in metrics_merge_experimental.groupby(['feature_experiment']):
                    ax = df.plot(ax=ax, kind='line',
                                      x='as_of_date_year',
                                      y='value',
                                      label=feature)
                metrics_merge[metrics_merge['experiment_type'] == 'All features']. \
                        groupby(['experiment_type']). \
                        plot(ax=ax,
                             kind='line',
                             x='as_of_date_year',
                             y='value',
                             label='All features')
                plt.title(str(metric).capitalize() +\
                          ' for selected model_groups in time.',
                          fontsize=fontsize)
                ax.tick_params(labelsize=16)
                ax.set_xlabel('Year of prediction (as_of_date)', fontsize=20)
                ax.set_ylabel(f'{str(metric)+str(param_type)+str(param)}',
                              fontsize=20)
                plt.xticks(model_metrics_filter.as_of_date_year.unique())
                plt.yticks(np.arange(0,1,0.1))
                legend=plt.legend(bbox_to_anchor=(1.05, 1),
                           loc=2,
                           borderaxespad=0.,
                           title='Experiment Type',
                           fontsize=fontsize)
                legend.get_title().set_fontsize('16')

            except TypeError:
                    print(f'''
                          Oops! model_metrics_pivot table is empty. Several problems
                          can be creating this error:
                          1. Check that {param_type}@{param} exists in the evaluations
                          table
                          2. Check that the metric {metric} is available to the
                          specified {param_type}@{param}.
                          3. You basline model can have different specifications.
                          Check those!
                          4. Check overlap between baseline dates and model dates.
                          The join is using the dates for doing these, and it's
                          possible that your timestamps differ.
                          ''')

    def _rank_corr_df(self,
                      model_pair,
                      top_n_features: 10,
                      corr_type=None,
                      param_type=None,
                      param=None):
        '''
        Calculates ranked correlations for ranked observations and features
        using the stats.spearmanr scipy module.
        Arguments:
            - model_pair (tuple): tuple with model_ids
              observations or features
            - top_n (int): number of rows to rank (top-k model)
        '''

        if corr_type not in ['predictions', 'features']:
            raise Exception(
                f'''Wrong type! Rank correlation is not available\n
                   for {type}. Try the following options:\n
                   predictions and features''')

        if corr_type == 'predictions':
            # Split df for each model_id
            model_1 = self.predictions[self.predictions['model_id'] == model_pair[0]]
            model_2 = self.predictions[self.predictions['model_id'] == model_pair[1]]

            # Slice df to take top-n observations
            top_model_1 = model_1[model_1[param_type] < param].set_index('entity_id')
            top_model_2 = model_2[model_2[param_type] < param].set_index('entity_id')

            # Merge df's by entity_id and calculate corr
            df_pair_merge = top_model_1.merge(top_model_2,
                                              how='inner',
                                              left_index=True,
                                              right_index=True,
                                              suffixes=['_1', '_2'])

            df_pair_filter = df_pair_merge.filter(regex='rank_abs*')
            rank_corr = spearmanr(df_pair_filter.iloc[:, 0], df_pair_filter.iloc[:, 1])

            # Return corr value (not p-value)
            return rank_corr[0]

        elif corr_type == 'features':
            # Split df for each model_id
            model_1 = \
                    self.feature_importances[self.feature_importances['model_id'] \
                                             == model_pair[0]]
            model_2 = \
                    self.feature_importances[self.feature_importances['model_id'] \
                                             == model_pair[1]]

            # Slice df to take top-n observations
            top_model_1 = model_1.sort_values('rank_abs', \
                                              axis=0)[:top_n_features].set_index('feature')
            top_model_2 = model_2.sort_values('rank_abs', \
                                              axis=0)[:top_n_features].set_index('feature')

            # Merge df's by entity_id and calculate corr
            df_pair_merge = top_model_1.merge(top_model_2,
                                              how='inner',
                                              left_index=True,
                                              right_index=True,
                                              suffixes=['_1', '_2'])

            df_pair_filter = df_pair_merge.filter(regex='rank_abs*')
            rank_corr = spearmanr(df_pair_filter.iloc[:, 0], df_pair_filter.iloc[:, 1])

            # Return corr value (not p-value)
            return rank_corr[0]
        else:
            pass

    def plot_ranked_correlation_preds(self,
                                       model_subset=None,
                                       temporal_comparison=False,
                                       figsize=(24, 10),
                                       fontsize=12,
                                       **kwargs):
        '''
        Plot ranked correlation between model_id's using the _rank_corr_df
        method. The plot will visualize the selected correlation matrix
        including all the models
        Arguments:
            - model_subset (list): subset to only include a subset of model_ids
            - corr_type (str): correlation type. Two options are available:
                features and predictions.
            - temporal_comparison (bool): Compare same prediction window models?
              Default is False
            - figzise (tuple): tuple with figure size. Default is (12, 16)
            - fontsize (int): Fontsize for plot labels and titles. Default is
              20
            - **kwargs: other parameters passed to the _rank_corr_df method
        '''
        if model_subset is None:
            model_subset = self.model_id
        models_to_use = []
        model_as_of_date = []
        for key, values in \
            self.same_time_models[['train_end_time', 'model_id_array']].iterrows():
            if values[1][0] in model_subset:
                models_to_use.append(values[1][0])
                model_as_of_date.append(values[0])
        model_subset = models_to_use

        if temporal_comparison == True:
            fig = plt.figure(figsize=figsize)
            for key, values in \
                self.same_time_models[['train_end_time', 'model_id_array']].iterrows():

                model_subset = values[1]
                model_as_of_date = values[0]

                # Calculate rank correlations for predictions
                corrs = [self._rank_corr_df(pair,
                         corr_type='predictions',
                         param=kwargs['param'],
                         param_type=kwargs['param_type'],
                         top_n_features=10
                         ) for pair in combinations(model_subset, 2)]

                # Store results in dataframe using tuples
                corr_matrix = pd.DataFrame(index=model_subset, columns=model_subset)
                for pair, corr in zip(combinations(model_subset, 2), corrs):
                    corr_matrix.loc[pair] = corr

                # Process data for plot: mask repeated tuples
                corr_matrix_t = corr_matrix.T
                mask = np.zeros_like(corr_matrix_t)
                mask[np.triu_indices_from(mask, k=1)] = True

                #fig, ax = plt.subplots(figsize=figsize)
                ax = fig.add_subplot(np.ceil(self.same_time_models.shape[0]/4), 4, key+1)
                ax.set_xlabel('Model Id', fontsize=fontsize)
                ax.set_ylabel('Model Id', fontsize=fontsize)
                plt.title(f'''Predictions Rank Correlation for
                          {kwargs['param_type']}@{kwargs['param']}
                          (date: {model_as_of_date})
                        ''', fontsize=fontsize)
                sns.heatmap(corr_matrix_t.fillna(1),
                            mask=mask,
                            vmax=1,
                            vmin=0,
                            cmap='YlGnBu',
                            annot=True,
                            square=True)

        else:
            corrs = [self._rank_corr_df(pair,
                                        corr_type='predictions',
                                        param=kwargs['param'],
                                        param_type=kwargs['param_type'],
                                        top_n_features=10
                                        ) for pair in combinations(model_subset, 2)]

            # Store results in dataframe using tuples
            corr_matrix = pd.DataFrame(index=model_subset, columns=model_subset)
            for pair, corr in zip(combinations(model_subset, 2), corrs):
                corr_matrix.loc[pair] = corr

            # Process data for plot: mask repeated tuples
            corr_matrix_t = corr_matrix.T
            mask = np.zeros_like(corr_matrix_t)
            mask[np.triu_indices_from(mask, k=1)] = True

            fig, ax = plt.subplots(figsize=figsize)
            ax.set_xlabel('Model Id', fontsize=fontsize)
            ax.set_ylabel('Model Id', fontsize=fontsize)
            plt.title(f'''Predictions Rank Correlation for
                      {kwargs['param_type']}@{kwargs['param']}
                     ''', fontsize=fontsize)
            sns.heatmap(corr_matrix_t.fillna(1),
                        mask=mask,
                        vmax=1,
                        vmin=0,
                        cmap='YlGnBu',
                        annot=True,
                        square=True)


    def plot_ranked_correlation_features(self,
                                         model_subset=None,
                                         temporal_comparison=False,
                                         figsize=(12, 16),
                                         fontsize=20,
                                         **kwargs):
        '''
        Plot ranked correlation between model_id's using the _rank_corr_df
        method. The plot will visualize the selected correlation matrix
        including all the models
        Arguments:
            - model_subset (list): subset to only include a subset of model_ids
            - corr_type (str): correlation type. Two options are available:
                features and predictions.
            - temporal_comarison (bool): Compare same prediction window models?
              Default is False
            - figzise (tuple): tuple with figure size. Default is (12, 16)
            - fontsize (int): Fontsize for plot labels and titles. Default is
              20
            - **kwargs: other parameters passed to the _rank_corr_df method
        '''

        if model_subset is None:
            model_subset = self.model_id
        models_to_use = []
        model_as_of_date = []
        for key, values in \
            self.same_time_models[['train_end_time', 'model_id_array']].iterrows():
            if values[1][0] in model_subset:
                models_to_use.append(values[1][0])
                model_as_of_date.append(values[0])
        model_subset = models_to_use

        if  temporal_comparison == True:

            # Calculate rank correlations for features
            corrs = [self._rank_corr_df(pair,
                                        corr_type='features',
                                        top_n_features = kwargs \
                                        ['top_n_features']
                                       ) for pair in combinations(model_subset, 2)]

            # Store results in dataframe using tuples
            corr_matrix = pd.DataFrame(index=model_subset, columns=model_subset)
            for pair, corr in zip(combinations(model_subset, 2), corrs):
                corr_matrix.loc[pair] = corr

            # Process data for plot: mask repeated tuples
            corr_matrix_t = corr_matrix.T
            mask = np.zeros_like(corr_matrix_t)
            mask[np.triu_indices_from(mask, k=1)] = True

            fig, ax = plt.subplots(figsize=figsize)
            ax.set_xlabel('Model Id', fontsize=fontsize)
            ax.set_ylabel('Model Id', fontsize=fontsize)
            plt.title(f'''Feature Rank Correlation for
                      Top-{kwargs['top_n_features']}
                     (date: {model_as_of_date})
                    ''', fontsize=fontsize)
            sns.heatmap(corr_matrix_t.fillna(1),
                        mask=mask,
                        vmax=1,
                        vmin=0,
                        cmap='YlGnBu',
                        annot=True,
                        square=True)

        else:
            corrs = [self._rank_corr_df(pair,
                                        corr_type='features',
                                        top_n_features=10
                                        ) for pair in combinations(model_subset, 2)]
            # Store results in dataframe using tuples
            corr_matrix = pd.DataFrame(index=model_subset, columns=model_subset)
            for pair, corr in zip(combinations(model_subset, 2), corrs):
                corr_matrix.loc[pair] = corr

            # Process data for plot: mask repeated tuples
            corr_matrix_t = corr_matrix.T
            mask = np.zeros_like(corr_matrix_t)
            mask[np.triu_indices_from(mask, k=1)] = True

            fig, ax = plt.subplots(figsize=figsize)
            ax.set_xlabel('Model Id', fontsize=fontsize)
            ax.set_ylabel('Model Id', fontsize=fontsize)
            plt.title(f'''Feature Rank Correlation for
                      Top-{kwargs['top_n_features']}
                    ''', fontsize=fontsize)
            sns.heatmap(corr_matrix_t.fillna(1),
                        mask=mask,
                        vmax=1,
                        vmin=0,
                        cmap='YlGnBu',
                        annot=True,
                        square=True)


    def plot_jaccard_preds(self,
                           param_type=None,
                           param=None,
                           model_subset=None,
                           temporal_comparison=False,
                           figsize=(24, 10),
                           fontsize=12):

        if model_subset is None:
            model_subset = self.model_id

        preds = self.predictions
        preds_filter = preds[preds['model_id'].isin(self.model_id)]

        if temporal_comparison == True:
            try:
                fig = plt.figure(figsize=figsize)
                for key, values in \
                self.same_time_models[['train_end_time', 'model_id_array']].iterrows():
                    preds_filter_group = \
                    preds_filter[preds_filter['model_id'].isin(values[1])]
                    # Filter predictions dataframe by individual dates
                    if param_type == 'rank_abs':
                        df_preds_date = preds_filter_group.copy()
                        df_preds_date['above_tresh'] = \
                                np.where(df_preds_date['rank_abs'] <= param, 1, 0)
                        df_sim_piv = df_preds_date.pivot(index='entity_id',
                                                         columns='model_id',
                                                         values='above_tresh')
                    elif param_type == 'rank_pct':
                        df_preds_date = preds_filter_group.copy()
                        df_preds_date['above_tresh'] = np.where(df_preds_date['rank_pct'] <= param, 1, 0)
                        df_preds_date['new_entity_id'] = df_preds_date['entity_id'].astype(str) + ":" + df_preds_date['as_of_date'].astype(str)
                        df_sim_piv = df_preds_date.pivot(index='new_entity_id',
                                                         columns='model_id',
                                                         values='above_tresh')
                    else:
                        raise AttributeError('''Error! You have to define a parameter type to
                                             set up a threshold
                                             ''')
                            # Calculate Jaccard Similarity for the selected models
                    res = pdist(df_sim_piv.T, 'jaccard')
                    df_jac = pd.DataFrame(1-squareform(res),
                                      index=preds_filter_group.model_id.unique(),
                                      columns=preds_filter_group.model_id.unique())
                    mask = np.zeros_like(df_jac)
                    mask[np.triu_indices_from(mask, k=1)] = True

                    # Plot matrix heatmap
                    ax = fig.add_subplot(np.ceil(self.same_time_models.shape[0]/4), 4, key+1)
                    ax.set_xlabel('Model Id', fontsize=fontsize)
                    ax.set_ylabel('Model Id', fontsize=fontsize)
                    plt.title(f'''(as_of_date:{values[0]})''', fontsize=fontsize)
                    sns.heatmap(df_jac,
                                mask=mask,
                                cmap='Greens',
                                vmin=0,
                                vmax=1,
                                annot=True,
                                linewidth=0.1)
            except ValueError:
                print(f'''
                      Temporal comparison can be only made for more than one
                      model group.
                     ''')

        else:
                # Call predicitons
                if param_type == 'rank_abs':
                    df_preds_date = preds_filter.copy()
                    df_preds_date['above_tresh'] = \
                            np.where(df_preds_date['rank_abs'] <= param, 1, 0)
                    df_sim_piv = df_preds_date.pivot(index='entity_id',
                                                     columns='model_id',
                                                     values='above_tresh')
                elif param_type == 'rank_pct':
                    df_preds_date = preds_filter.copy()
                    df_preds_date['above_tresh'] = \
                            np.where(df_preds_date['rank_pct'] <= param, 1, 0)
                    df_preds_date['new_entity_id'] = df_preds_date['entity_id'].astype(str) + ":" + df_preds_date['as_of_date'].astype(str)
                    df_sim_piv = df_preds_date.pivot(index='new_entity_id',
                                                     columns='model_id',
                                                     values='above_tresh')
                else:
                    raise AttributeError('''Error! You have to define a parameter type to
                                         set up a threshold
                                         ''')

                # Calculate Jaccard Similarity for the selected models
                res = pdist(df_sim_piv[model_subset].T, 'jaccard')
                df_jac = pd.DataFrame(1-squareform(res),
                                      index=model_subset,
                                      columns=model_subset)
                mask = np.zeros_like(df_jac)
                mask[np.triu_indices_from(mask, k=1)] = True

                # Plot matrix heatmap
                fig, ax = plt.subplots(figsize=figsize)
                ax.set_xlabel('Model Id', fontsize=fontsize)
                ax.set_ylabel('Model Id', fontsize=fontsize)
                plt.title('Jaccard Similarity Matrix Plot', fontsize=fontsize)
                sns.heatmap(df_jac,
                            mask=mask,
                            cmap='Greens',
                            vmin=0,
                            vmax=1,
                            annot=True,
                            linewidth=0.1)


    def plot_jaccard_features(self,
                              top_n_features=10,
                              model_subset=None,
                              temporal_comparison=False,
                              figsize=(30, 10),
                              fontsize=12):

        if model_subset is None:
            model_subset = self.model_id

        f_importances = self.feature_importances
        f_importances_filter = \
        f_importances[f_importances['model_id'].isin(model_subset)]

        if temporal_comparison == True:
            try:
                fig = plt.figure(figsize=figsize)
                for key, values in \
                self.same_time_models[['train_end_time', 'model_id_array']].iterrows():
                    f_imp_filter_group = \
                    f_importances_filter[f_importances_filter['model_id'].isin(values[1])]
                    
                    if top_n_features is not None:
                        f_imp_date = f_imp_filter_group.copy()
                        f_imp_date_filter = \
                                f_imp_filter_group.sort_values('rank_abs')
                        f_imp_date_filter_top = \
                                f_imp_date_filter[f_imp_date_filter['rank_abs']
                                                 <= top_n_features]

                        df_sim_piv = f_imp_date_filter_top.pivot(index='feature',
                                                                 columns='model_id',
                                                                 values='rank_abs')
                    else:
                        raise AttributeError('''Error! You have to define a top_n features to
                                             set up a threshold
                                             ''')

                    # Calculate Jaccard Similarity for the selected models
                    res = pdist(df_sim_piv.T, 'jaccard')
                    df_jac = pd.DataFrame(1-squareform(res),
                                      index=values[1],
                                      columns=values[1])
                    mask = np.zeros_like(df_jac)
                    mask[np.triu_indices_from(mask, k=1)] = True

                    # Plot matrix heatmap
                    ax = fig.add_subplot(np.ceil(self.same_time_models.shape[0]/4), 4, key+1)
                    ax.set_xlabel('Model Id', fontsize=fontsize)
                    ax.set_ylabel('Model Id', fontsize=fontsize)
                    plt.title(f'''(as_of_date:{values[0]})''', fontsize=fontsize)
                    sns.heatmap(df_jac,
                                mask=mask,
                                cmap='Greens',
                                vmin=0,
                                vmax=1,
                                annot=True,
                                linewidth=0.1)
            except ValueError:
                print(f'''
                      Temporal comparison can be only made for more than one
                      model group.
                     ''')

        else:
                # Call predicitons
                if top_n_features is not None:
                    f_importances_filter_all = f_importances_filter.copy()
                    f_importance_filter_all_rank = \
                            f_importances_filter_all.sort_values('rank_abs')
                    f_importance_filter_all_rank_top = \
                            f_importance_filter_all_rank[f_importance_filter_all_rank['rank_abs']
                                                        <= top_n_features]

                    df_sim_piv = \
                    f_importance_filter_all_rank_top.pivot(index='feature',
                                                           columns='model_id',
                                                           values='rank_abs')
                else:
                    raise AttributeError('''Error! You have to define a parameter type to
                                         set up a threshold
                                         ''')

                # Calculate Jaccard Similarity for the selected models
                res = pdist(df_sim_piv[model_subset].T, 'jaccard')
                df_jac = pd.DataFrame(1-squareform(res),
                                      index=model_subset,
                                      columns=model_subset)
                mask = np.zeros_like(df_jac)
                mask[np.triu_indices_from(mask, k=1)] = True

                # Plot matrix heatmap
                fig, ax = plt.subplots(figsize=figsize)
                ax.set_xlabel('Model Id', fontsize=fontsize)
                ax.set_ylabel('Model Id', fontsize=fontsize)
                plt.title('Jaccard Similarity Matrix Plot', fontsize=fontsize)
                sns.heatmap(df_jac,
                            mask=mask,
                            cmap='Greens',
                            vmin=0,
                            vmax=1,
                            annot=True,
                            linewidth=0.1)


    def _plot_preds_compare_score_dist(self,
                                       m0,
                                       m1,
                                       df_preds_date,
                                       colors=['blue', 'orange'],
                                       bins=np.arange(0,1.01,0.01)):

        '''
        Plotting function for comparing prediction distributions across models.
        This function takes two model_ids with predictions in the same
        prediction window and shows the relative distribution of
        the first model top-k in the second model score distribution.

        This function is meant to be used as a helper function of
        plot_preds_comparisons.

        Arguments:
            - m0, m1: (int) model_id
            - df_preds_date: (dataframe) predictions dataframe
            - colors: (str) color strings. Defaults are blue and orange
            - bins: (np array) number of bins to pass to the seaborn histogram
            plotting function

        Returns:
            matplotlib plot object
        '''

        df_preds_m0 = df_preds_date[df_preds_date['model_id']==m0]
        df_preds_m1 = df_preds_date[df_preds_date['model_id']==m1]

        sns.distplot(df_preds_m0[df_preds_m0['above_tresh']==0]['score'],
                     kde=False,
                     bins=bins,
                     color='grey',
                     label="model " + str(m0) + " predicted label = 0")
        sns.distplot(df_preds_m0[df_preds_m0['above_tresh']==1]['score'],
                     kde=False,
                     bins=bins,
                     color=colors[1],
                     label="model " + str(m0) + " predicted label = 1")

        df_alt_model_scores = \
                pd.merge(df_preds_m0, df_preds_m1[df_preds_m1.above_tresh==1][['entity_id', 'as_of_date']])

        sns.distplot(df_alt_model_scores['score'],
                     kde=False,
                     bins=bins,
                     color=colors[0],
                     label="model " + str(m1) + " predicted label = 1")

        plt.xlabel("Scores from model " + str(m0))
        plt.legend()

    def _plot_preds_compare_rank(self,
                                 m0,
                                 m1,
                                 df_preds_date,
                                 colors=['black'],
                                 show_tp_fp=False,
                                 bins = np.arange(0,110,10)):
        '''
        Plot predictions rank comparison for two selected models.

        This function will rank the predictions from one model into the decile
        distribution of the second one. This function is meant to be used as a
        part of the plot_preds_comparison function.

        Arguments:
            - m0, m1: (int) model_ids to compare, only two.
            - df_preds_date: (dataframe) predictions dataframe
            - colors: (str) color string. Default is black
            - show_tp_tn: (bool) Plot true positive and true negatives in the
                           rank distribution plot. Default is False
            - bins: (np array) Number of bins to pass to the seaborn
            histogram plot function.
        '''

        df_preds_m0 = df_preds_date[df_preds_date['model_id']==m0]
        df_preds_m1 = df_preds_date[df_preds_date['model_id']==m1]
        df_alt_model_rank = \
                pd.merge(df_preds_m0, df_preds_m1[df_preds_m1.above_tresh==1][['entity_id', 'as_of_date']])

        if show_tp_fp:
            sns.distplot(df_alt_model_rank[df_alt_model_rank['label_value']==0]['rank_pct'],
                         kde=False,
                         bins=bins,
                         hist=True,
                         color=colors[0],
                         label="false positives")
            sns.distplot(df_alt_model_rank[df_alt_model_rank['label_value']==1]['rank_pct'],
                         kde=False,
                         bins=bins,
                         hist=True,
                         color=colors[1],
                         label="true positives")
            plt.legend()
        else:
            sns.distplot(df_alt_model_rank['rank_pct'],
                         kde=False,
                         bins=bins,
                         hist=True,
                         color=colors[0])
        plt.xlabel("Percentile Rank in model " + str(m0))
        plt.title("model "+str(m1)+" predicted label = 1")
        plt.xticks(bins)

    def plot_preds_comparison(self,
                              param_type=None,
                              param=None,
                              model_subset=None,
                              figsize=(28, 16),
                              fontsize=12):
        '''
        Plot predictor distribution comparison (distribution and rank)

        This function compares the predictions of all models, or a subset passed to
        model_subset. To compare predictions, the function will show the
        relative position of the score distribution of the top-k of one of the
        models into another model.

        Also, to compare how "off" can predictions be, the function will plot
        the rank position of the predictions of one model in to another. The
        plot will show the decile position of one model into the other.

        Arguments:
            - param_type: (str) parameter type (i.e. 'rank_abs', 'rank_pct')
            - param: (int) parameter threshold
            - model_subset: (list) list of model_ids to compare. Default is
            none, and the function will take all the models in self.model_id
            - figsize, fontsize: aesthetics for plots.
        '''

        if model_subset is None:
            model_subset = self.model_id

        preds = self.predictions
        preds_filter = preds[preds['model_id'].isin(self.model_id)]

        fig = plt.figure(figsize=figsize)
        try:
            for key, values in \
                self.same_time_models[['train_end_time', 'model_id_array']].iterrows():
                preds_filter_group = \
                                     preds_filter[preds_filter['model_id'].isin(values[1])]
                # Filter predictions dataframe by individual dates
                df_preds_date = preds_filter_group.copy()
                if param_type == 'rank_abs':
                    df_preds_date['above_tresh'] = \
                        np.where(df_preds_date['rank_abs'] <= param, 1, 0)
                elif param_type == 'rank_pct':
                    df_preds_date['above_tresh'] = \
                        np.where(df_preds_date['rank_pct'] <= param, 1, 0)

                else:
                    raise AttributeError('''Error! You have to define a parameter type to
                    set up a threshold
                    ''')

                sns.set_style('whitegrid')
                sns.set_context("poster", font_scale=1.25, rc={"lines.linewidth": 2.25,"lines.markersize":12})
                plt.clf()
                fig = plt.figure(figsize=figsize)
                for pair in itertools.combinations(values['model_id_array'], 2):
                    m0 = pair[0]
                    m1 = pair[1]
                    colors = {m0: 'blue', m1: 'orange'}
                    ax1 = plt.subplot(231)
                    self._plot_preds_compare_score_dist(pair[0], pair[1],
                                                       df_preds_date,
                                                       colors=[colors[m0], colors[m1]])
                    ax1 = plt.subplot(234)
                    self._plot_preds_compare_score_dist(pair[1],
                                                       pair[0],
                                                       df_preds_date,
                                                       colors=[colors[m1], colors[m0]])
                    ax1 = plt.subplot(232)
                    self._plot_preds_compare_rank(pair[0],
                                                 pair[1],
                                                 df_preds_date,
                                                 colors=[colors[m0]])
                    ax1 = plt.subplot(235)
                    self._plot_preds_compare_rank(pair[1],
                                                 pair[0],
                                                 df_preds_date,
                                                 colors=[colors[m1]])
                    ax1 = plt.subplot(233)
                    self._plot_preds_compare_rank(pair[0],
                                                 pair[1],
                                                 df_preds_date,
                                                 show_tp_fp=True,
                                                 colors=['lightblue', 'darkblue'])
                    ax1 = plt.subplot(236)
                    self._plot_preds_compare_rank(pair[1],
                                                 pair[0],
                                                 df_preds_date,
                                                 show_tp_fp=True,
                                                 colors=['khaki', 'darkorange'])
                plt.tight_layout()
                fig.suptitle(values['train_end_time'])
                plt.show()
        except ValueError:
            print(f'''
            Temporal comparison can be only made for more than one
            model group.
            ''')



# -*- coding: utf-8 -*-
from itertools import chain
import sqlalchemy.sql.expression as ex
from descriptors import cachedproperty

from .sql import make_sql_clause
from .collate import Aggregation


class SpacetimeAggregation(Aggregation):
    def __init__(
        self,
        aggregates,
        groups,
        intervals,
        from_obj,
        dates,
        state_table,
        state_group=None,
        prefix=None,
        suffix=None,
        schema=None,
        date_column=None,
        output_date_column=None,
        input_min_date=None,
        join_with_cohort_table=False,
    ):
        """
        Args:
            intervals: the intervals to aggregate over. either a list of
                datetime intervals, e.g. ["1 month", "1 year"], or
                a dictionary of group : intervals pairs where
                group is a group in groups and intervals is a collection
                of datetime intervals, e.g. {"address_id": ["1 month", "1 year]}
            dates: list of PostgreSQL date strings,
                e.g. ["2012-01-01", "2013-01-01"]
            state_table: schema.table to query for valid state_group/date combinations
            state_group: the group level found in the state table (e.g., "entity_id")
            date_column: name of date column in from_obj, defaults to "date"
            output_date_column: name of date column in aggregated output, defaults to "date"
            input_min_date: minimum date for which rows shall be included, defaults
                to no absolute time restrictions on the minimum date of included rows

        For all other arguments see collate.Aggregation
        """
        Aggregation.__init__(
            self,
            aggregates=aggregates,
            from_obj=from_obj,
            groups=groups,
            state_table=state_table,
            state_group=state_group,
            prefix=prefix,
            suffix=suffix,
            schema=schema,
        )

        if isinstance(intervals, dict):
            self.intervals = intervals
        else:
            self.intervals = {g: intervals for g in self.groups}
        self.dates = dates
        self.date_column = date_column if date_column else "date"
        self.output_date_column = output_date_column if output_date_column else "date"
        self.input_min_date = input_min_date
        self.join_with_cohort_table = join_with_cohort_table

    def _state_table_sub(self):
        """Helper function to ensure we only include state table records
        in our set of input dates and after the input_min_date.
        """
        datestr = ", ".join(["'%s'::date" % dt for dt in self.dates])
        mindtstr = (
            " AND %s >= '%s'::date" % (self.output_date_column, self.input_min_date)
            if self.input_min_date is not None
            else ""
        )
        return """(
        SELECT *
        FROM {st}
        WHERE {datecol} IN ({datestr})
        {mindtstr})""".format(
            st=self.state_table,
            datecol=self.output_date_column,
            datestr=datestr,
            mindtstr=mindtstr,
        )

    @cachedproperty
    def colname_aggregate_lookup(self):
        """A reverse lookup from column name to the source collate.Aggregate

        Will error if the Aggregation contains duplicate column names
        """
        lookup = {}
        for group, groupby in self.groups.items():
            intervals = self.intervals[group]
            for interval in intervals:
                date = self.dates[0]
                for agg in self.aggregates:
                    for col in self._cols_for_aggregate(agg, group, interval, date):
                        if col.name in lookup:
                            raise ValueError("Duplicate feature column name found: ", col.name)
                        lookup[col.name] = agg

        return lookup
            
    def _col_prefix(self, group, interval):
        """
        Helper for creating a column prefix for the group
            group: group clause, for naming columns
            interval: SQL time interval string, or "all"
        Returns: string for a common column prefix for columns in that group and interval
        """
        return "{prefix}_{group}_{interval}_".format(
            prefix=self.prefix, interval=interval, group=group
        )

    def _cols_for_aggregate(self, agg, group, interval, date):
        """
        Helper for getting the sql for a particular aggregate
        Args:
            agg: collate.Aggregate
            interval: SQL time interval string, or "all"
            date: SQL date string
            group: group clause, for naming columns
        Returns: collection of aggregate column SQL strings
        """
        if interval != "all":
            when = "{date_column} >= '{date}'::date - interval '{interval}'".format(
                interval=interval, date=date, date_column=self.date_column
            )
        else:
            when = None
        return agg.get_columns(
            when,
            self._col_prefix(group, interval),
            format_kwargs={"collate_date": date, "collate_interval": interval},
        )

    def _get_aggregates_sql(self, interval, date, group):
        """
        Helper for getting aggregates sql
        Args:
            interval: SQL time interval string, or "all"
            date: SQL date string
            group: group clause, for naming columns
        Returns: collection of aggregate column SQL strings
        """
        return chain(
            *[
                self._cols_for_aggregate(agg, group, interval, date)
                for agg in self.aggregates
            ]
        )

    def get_selects(self):
        """
        Constructs select queries for this aggregation

        Returns: a dictionary of group : queries pairs where
            group are the same keys as groups
            queries is a list of Select queries, one for each date in dates
        """
        queries = {}

        for group, groupby in self.groups.items():
            intervals = self.intervals[group]
            queries[group] = []
            for date in self.dates:
                columns = [
                    make_sql_clause(groupby, ex.text),
                    ex.literal_column("'%s'::date" % date).label(
                        self.output_date_column
                    ),
                ]
                columns += list(
                    chain(
                        *[self._get_aggregates_sql(i, date, group) for i in intervals]
                    )
                )

                gb_clause = make_sql_clause(groupby, ex.literal_column)
                if self.join_with_cohort_table:
                    from_obj = ex.text(
                        f"(select from_obj.* from ("
                        f"(select * from {self.from_obj}) from_obj join {self.state_table} cohort on ( "
                        "cohort.entity_id = from_obj.entity_id and "
                        f"cohort.{self.output_date_column} = '{date}'::date)"
                        ")) cohorted_from_obj")
                else:
                    from_obj = self.from_obj
                query = ex.select(columns=columns, from_obj=make_sql_clause(from_obj, ex.text)).group_by(
                    gb_clause
                )
                query = query.where(self.where(date, intervals))

                queries[group].append(query)

        return queries

    def get_imputation_rules(self):
        """
        Constructs a dictionary to lookup an imputation rule from an associated
        column name.

        Returns: a dictionary of column : imputation_rule pairs
        """
        imprules = {}
        for group, groupby in self.groups.items():
            for interval in self.intervals[group]:
                prefix = "{prefix}_{group}_{interval}_".format(
                    prefix=self.prefix, interval=interval, group=group
                )
                for a in self.aggregates:
                    imprules.update(a.column_imputation_lookup(prefix=prefix))
        return imprules

    def where(self, date, intervals):
        """
        Generates a WHERE clause
        Args:
            date: the end date
            intervals: intervals

        Returns: a clause for filtering the from_obj to be between the date and
            the greatest interval
        """
        # upper bound
        w = "{date_column} < '{date}'".format(date_column=self.date_column, date=date)

        # lower bound (if possible)
        if "all" not in intervals:
            greatest = "greatest(%s)" % str.join(
                ",", ["interval '%s'" % i for i in intervals]
            )
            min_date = "'{date}'::date - {greatest}".format(
                date=date, greatest=greatest
            )
            w += "AND {date_column} >= {min_date}".format(
                date_column=self.date_column, min_date=min_date
            )
        if self.input_min_date is not None:
            w += "AND {date_column} >= '{bot}'::date".format(
                date_column=self.date_column, bot=self.input_min_date
            )
        return ex.text(w)

    def get_indexes(self):
        """
        Generate create index queries for this aggregation

        Returns: a dictionary of group : index pairs where
            group are the same keys as groups
            index is a raw create index query for the corresponding table
        """
        return {
            group: "CREATE INDEX ON %s (%s, %s);"
            % (self.get_table_name(group), groupby, self.output_date_column)
            for group, groupby in self.groups.items()
        }

    def get_join_table(self):
        """
        Generates a join table, consisting of an entry for each combination of
        groups and dates in the from_obj
        """
        groups = [make_sql_clause(group, ex.text) for group in self.groups.values()]
        intervals = list(set(chain(*self.intervals.values())))

        queries = []
        for date in self.dates:
            columns = groups + [
                ex.literal_column("'%s'::date" % date).label(self.output_date_column)
            ]
            queries.append(
                ex.select(columns, from_obj=make_sql_clause(self.from_obj, ex.text))
                .where(self.where(date, intervals))
                .group_by(*groups)
            )

        return str.join("\nUNION ALL\n", map(str, queries))

    def get_create(self, join_table=None):
        """
        Generate a single aggregation table creation query by joining
            together the results of get_creates()
        Returns: a CREATE TABLE AS query
        """
        if not join_table:
            join_table = "(%s) t1" % self.get_join_table()
        query = "SELECT * FROM %s\n" % join_table
        for group, groupby in self.groups.items():
            query += " LEFT JOIN %s USING (%s, %s)" % (
                self.get_table_name(group),
                groupby,
                self.output_date_column,
            )

        return "CREATE TABLE %s AS (%s);" % (self.get_table_name(), query)

    def validate(self, conn):
        """
        SpacetimeAggregations ensure that no intervals extend beyond the absolute
        minimum time.
        """
        if self.input_min_date is not None:
            all_intervals = set(*self.intervals.values())
            for date in self.dates:
                for interval in all_intervals:
                    if interval == "all":
                        continue
                    # This could be done more efficiently all at once, but doing
                    # it this way allows for nicer error messages.
                    r = conn.execute(
                        "select ('%s'::date - '%s'::interval) < '%s'::date"
                        % (date, interval, self.input_min_date)
                    )
                    if r.fetchone()[0]:
                        raise ValueError(
                            "date '%s' - '%s' is before input_min_date ('%s')"
                            % (date, interval, self.input_min_date)
                        )
                    r.close()
        for date in self.dates:
            r = conn.execute(
                "select count(*) from %s where %s = '%s'::date"
                % (self.state_table, self.output_date_column, date)
            )
            if r.fetchone()[0] == 0:
                raise ValueError(
                    "date '%s' is not present in states table ('%s')"
                    % (date, self.state_table)
                )
            r.close()

    def find_nulls(self, imputed=False):
        """
        Generate query to count number of nulls in each column in the aggregation table

        Returns: a SQL SELECT statement
        """
        query_template = """
            SELECT {cols}
            FROM {state_tbl} t1
            LEFT JOIN {aggs_tbl} t2 USING({group}, {date_col})
            """
        cols_sql = ",\n".join(
            [
                """SUM(CASE WHEN "{col}" IS NULL THEN 1 ELSE 0 END) AS "{col}" """.format(
                    col=column
                )
                for column in self.get_imputation_rules().keys()
            ]
        )

        return query_template.format(
            cols=cols_sql,
            state_tbl=self._state_table_sub(),
            aggs_tbl=self.get_table_name(imputed=imputed),
            group=self.state_group,
            date_col=self.output_date_column,
        )

    def get_impute_create(self, impute_cols, nonimpute_cols):
        """
        Generates the CREATE TABLE query for the aggregation table with imputation.

        Args:
            impute_cols: a list of column names with null values
            nonimpute_cols: a list of column names without null values

        Returns: a CREATE TABLE AS query
        """

        # key columns and date column
        query = "SELECT %s, %s" % (
            ", ".join(map(str, self.groups.values())),
            self.output_date_column,
        )

        # columns with imputation filling as needed
        query += self._get_impute_select(
            impute_cols, nonimpute_cols, partitionby=self.output_date_column
        )

        # imputation starts from the state table and left joins into the aggregation table
        query += "\nFROM %s t1" % self._state_table_sub()
        query += "\nLEFT JOIN %s t2 USING(%s, %s)" % (
            self.get_table_name(),
            self.state_group,
            self.output_date_column,
        )

        return "CREATE TABLE %s AS (%s)" % (self.get_table_name(imputed=True), query)



"""Changing column names and removing redundancies in table names

Revision ID: 2446a931de7a
Revises: 89a8ce240bae
Create Date: 2018-05-24 17:07:20.567789

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "2446a931de7a"
down_revision = "89a8ce240bae"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.execute(
        "ALTER TABLE test_results.test_evaluations RENAME TO evaluations;"
        + "ALTER TABLE test_results.test_predictions RENAME TO predictions;"
        + "ALTER TABLE train_results.train_evaluations RENAME TO evaluations;"
        + "ALTER TABLE train_results.train_predictions RENAME TO predictions;"
    )

    op.alter_column(
        "matrices",
        "n_examples",
        new_column_name="num_observations",
        schema="model_metadata",
    )
    op.alter_column(
        "model_groups",
        "model_parameters",
        new_column_name="hyperparameters",
        schema="model_metadata",
    )
    op.alter_column(
        "models",
        "model_parameters",
        new_column_name="hyperparameters",
        schema="model_metadata",
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.execute(
        "ALTER TABLE test_results.evaluations RENAME TO test_evaluations;"
        + "ALTER TABLE test_results.predictions RENAME TO test_predictions;"
        + "ALTER TABLE train_results.evaluations RENAME TO train_evaluations;"
        + "ALTER TABLE train_results.predictions RENAME TO train_predictions;"
    )

    op.alter_column(
        "matrices",
        "num_observations",
        new_column_name="n_examples",
        schema="model_metadata",
    )
    op.alter_column(
        "model_groups",
        "hyperparameters",
        new_column_name="model_parameters",
        schema="model_metadata",
    )
    op.alter_column(
        "models",
        "hyperparameters",
        new_column_name="model_parameters",
        schema="model_metadata",
    )

    # ### end Alembic commands ###



from datetime import datetime, date

import testing.postgresql
from sqlalchemy.engine import create_engine
from unittest.mock import MagicMock

from triage.component.catwalk.protected_groups_generators import ProtectedGroupsGenerator


def create_demographics_table(db_engine, data):
    db_engine.execute(
        """drop table if exists demographics;
        create table demographics (person_id int, event_date date, race text, sex text)
        """
    )
    for event in data:
        db_engine.execute(
            "insert into demographics values (%s, %s, %s, %s)", event
        )


def create_cohort_table(db_engine, data):
    db_engine.execute(
        "create table cohort_abcdef (entity_id int, as_of_date timestamp)"
    )
    for event in data:
        db_engine.execute(
            "insert into cohort_abcdef values (%s, %s)", event
        )


def default_demographics():
    return [
        (1, datetime(2015, 12, 30), 'aa', 'male'),
        (1, datetime(2016, 2, 1), 'aa', 'male'),
        (1, datetime(2016, 3, 1), 'aa', 'female'),
        (2, datetime(2015, 12, 30), 'wh', 'male'),
        (2, datetime(2016, 3, 1), 'wh', 'male'),
        (3, datetime(2015, 12, 30), 'aa', 'male'),
        (3, datetime(2016, 3, 1), 'aa', 'male'),
        (5, datetime(2016, 2, 1), 'wh', 'female'),
        (5, datetime(2016, 3, 1), 'wh', 'female'),
    ]


def default_cohort():
    return [
        (1, datetime(2016, 1, 1)),
        (1, datetime(2016, 3, 1)),
        (1, datetime(2016, 4, 1)),
        (2, datetime(2016, 1, 1)),
        (2, datetime(2016, 3, 1)),
        (2, datetime(2016, 4, 1)),
        (3, datetime(2016, 1, 1)),
        (3, datetime(2016, 3, 1)),
        (3, datetime(2016, 4, 1)),
        (4, datetime(2016, 1, 1)),
        (4, datetime(2016, 3, 1)),
        (4, datetime(2016, 4, 1)),
        (5, datetime(2016, 1, 1)),
        (5, datetime(2016, 3, 1)),
        (5, datetime(2016, 4, 1)),
    ]


def assert_data(table_generator):
    expected_output = [
        (1, date(2016, 1, 1), 'aa', 'male', 'abcdef'),
        (1, date(2016, 3, 1), 'aa', 'male', 'abcdef'),
        (1, date(2016, 4, 1), 'aa', 'female', 'abcdef'),
        (2, date(2016, 1, 1), 'wh', 'male', 'abcdef'),
        (2, date(2016, 3, 1), 'wh', 'male', 'abcdef'),
        (2, date(2016, 4, 1), 'wh', 'male', 'abcdef'),
        (3, date(2016, 1, 1), 'aa', 'male', 'abcdef'),
        (3, date(2016, 3, 1), 'aa', 'male', 'abcdef'),
        (3, date(2016, 4, 1), 'aa', 'male', 'abcdef'),
        (4, date(2016, 1, 1), None, None, 'abcdef'),
        (4, date(2016, 3, 1), None, None, 'abcdef'),
        (4, date(2016, 4, 1), None, None, 'abcdef'),
        (5, date(2016, 1, 1), None, None, 'abcdef'),
        (5, date(2016, 3, 1), 'wh', 'female', 'abcdef'),
        (5, date(2016, 4, 1), 'wh', 'female', 'abcdef'),
    ]
    results = list(
        table_generator.db_engine.execute(
            f"""
            select entity_id, as_of_date, race, sex, cohort_hash
            from {table_generator.protected_groups_table_name}
            order by entity_id, as_of_date
        """
        )
    )
    assert results == expected_output


def test_protected_groups_generator_replace():
    demographics_data = default_demographics()
    cohort_data = default_cohort()
    with testing.postgresql.Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        create_demographics_table(engine, demographics_data)
        create_cohort_table(engine, cohort_data)
        table_generator = ProtectedGroupsGenerator(
            from_obj="demographics",
            attribute_columns=['race', 'sex'],
            entity_id_column="person_id",
            knowledge_date_column="event_date",
            db_engine=engine,
            protected_groups_table_name="protected_groups_abcdef",
            replace=True
        )
        as_of_dates = [
            datetime(2016, 1, 1),
            datetime(2016, 3, 1),
            datetime(2016, 4, 1),
        ]
        table_generator.generate_all_dates(
            as_of_dates,
            cohort_table_name='cohort_abcdef',
            cohort_hash='abcdef'
        )
        assert_data(table_generator)

        table_generator.generate_all_dates(
            as_of_dates,
            cohort_table_name='cohort_abcdef',
            cohort_hash='abcdef'
        )
        assert_data(table_generator)


def test_protected_groups_generator_noreplace():
    demographics_data = default_demographics()
    cohort_data = default_cohort()
    with testing.postgresql.Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        create_demographics_table(engine, demographics_data)
        create_cohort_table(engine, cohort_data)
        table_generator = ProtectedGroupsGenerator(
            from_obj="demographics",
            attribute_columns=['race', 'sex'],
            entity_id_column="person_id",
            knowledge_date_column="event_date",
            db_engine=engine,
            protected_groups_table_name="protected_groups_abcdef",
            replace=False
        )
        as_of_dates = [
            datetime(2016, 1, 1),
            datetime(2016, 3, 1),
            datetime(2016, 4, 1),
        ]
        table_generator.generate_all_dates(
            as_of_dates,
            cohort_table_name='cohort_abcdef',
            cohort_hash='abcdef'
        )
        assert_data(table_generator)
        table_generator.generate = MagicMock()
        table_generator.generate_all_dates(
            as_of_dates,
            cohort_table_name='cohort_abcdef',
            cohort_hash='abcdef'
        )
        table_generator.generate.assert_not_called()
        assert_data(table_generator)



import pathlib
import subprocess
from sqlalchemy import create_engine


DATA_NAME = "food_inspections_subset.csv"
DATA_PATH = pathlib.Path(__file__).with_name(DATA_NAME)


def handler(database):
    engine = create_engine(database.url())
    connection = engine.connect()
    try:
        load_data(connection)
    finally:
        connection.close()


def load_data(connection):
    connection.execute("DROP TABLE IF EXISTS food_inspections")
    subprocess.run(
        [
            "csvsql",
            "-v",
            "--no-constraints",
            "--tables",
            "food_inspections",
            "--insert",
            "--db",
            str(connection.engine.url),
            str(DATA_PATH),
        ],
        check=True,
    )
    connection.execute("CREATE INDEX ON food_inspections(license_no, inspection_date)")

    # create a state table for license/date
    connection.execute("DROP TABLE IF EXISTS inspection_states")
    connection.execute(
        """\
        CREATE TABLE inspection_states AS (
            SELECT license_no, date
            FROM (SELECT DISTINCT license_no FROM food_inspections) a
            CROSS JOIN (SELECT DISTINCT inspection_date as date FROM food_inspections) b
        )"""
    )
    connection.execute("CREATE INDEX ON inspection_states(license_no, date)")

    # create an alternate state table with a different date column
    connection.execute("DROP TABLE IF EXISTS inspection_states_diff_colname")
    connection.execute(
        """\
        CREATE TABLE inspection_states_diff_colname
        AS select license_no, date as aggregation_date
        FROM inspection_states
        """
    )
    connection.execute(
        """\
        CREATE INDEX ON
        inspection_states_diff_colname(license_no, aggregation_date)
        """
    )

    # create a state table for licenseo only
    connection.execute("DROP TABLE IF EXISTS all_licenses")
    connection.execute(
        """\
        CREATE TABLE all_licenses AS (
            SELECT DISTINCT license_no FROM food_inspections
        )"""
    )
    connection.execute("CREATE INDEX ON all_licenses(license_no)")



from triage.component import results_schema
from alembic import command, script
import pytest


def test_upgrade_if_clean_upgrades_if_clean(db_engine):
    results_schema.upgrade_if_clean(db_engine.url)
    db_version = db_engine.execute("select version_num from results_schema_versions").scalar()
    alembic_cfg = results_schema.alembic_config(db_engine.url)
    assert db_version == script.ScriptDirectory.from_config(alembic_cfg).get_current_head()


def test_upgrade_if_clean_does_not_upgrade_if_not_clean(db_engine):
    command.upgrade(results_schema.alembic_config(dburl=db_engine.url), "head")
    command.downgrade(results_schema.alembic_config(dburl=db_engine.url), "-1")
    with pytest.raises(ValueError):
        results_schema.upgrade_if_clean(db_engine.url)



class BaseRules:
    def __init__(self):
        self._metric = None
        self._parameter = None
        self.shared_parameters = []
        self.selection_rules = []

    def _does_parameters_exist(self, params_dict):
        return params_dict in self.shared_parameters

    def _does_selection_rule_exisit(self, rule_dict):
        return rule_dict in self.selection_rules

    def _append(self, params_dict, rule_dict):
        if not self._does_parameters_exist(params_dict):
            self.shared_parameters.append(params_dict)
        if not self._does_selection_rule_exisit(rule_dict):
            self.selection_rules.append(rule_dict)

    def create(self):
        return [
            {
                "shared_parameters": self.shared_parameters,
                "selection_rules": self.selection_rules,
            }
        ]


class SimpleRuleMaker(BaseRules):

    """
    Holds methods that generate parameter grids for selection rules that
    evaluate the performance of a model group in terms of a single metric.
    These include:

    - [best_current_value][triage.component.audition.selection_rules.best_current_value]
    - [best_average_value][triage.component.audition.selection_rules.best_average_value]
    - [lowest_metric_variance][triage.component.audition.selection_rules.lowest_metric_variance]
    - [most_frequent_best_dist][triage.component.audition.selection_rules.most_frequent_best_dist]
    - [best_avg_var_penalized][triage.component.audition.selection_rules.best_avg_var_penalized]
    - [best_avg_recency_weight][triage.component.audition.selection_rules.best_avg_recency_weight]
    """

    def add_rule_best_current_value(self, metric=None, parameter=None, n=1):
        if metric is not None:
            self._metric = metric
        if parameter is not None:
            self._parameter = parameter
        params_dict = {"metric": self._metric, "parameter": self._parameter}
        rule_dict = {"name": "best_current_value", "n": n}
        self._append(params_dict, rule_dict)
        return self.create()

    def add_rule_best_average_value(self, metric=None, parameter=None, n=1):
        if metric is not None:
            self._metric = metric
        if parameter is not None:
            self._parameter = parameter
        params_dict = {"metric": self._metric, "parameter": self._parameter}
        rule_dict = {"name": "best_average_value", "n": n}
        self._append(params_dict, rule_dict)
        return self.create()

    def add_rule_lowest_metric_variance(self, metric=None, parameter=None, n=1):
        if metric is not None:
            self._metric = metric
        if parameter is not None:
            self._parameter = parameter
        params_dict = {"metric": self._metric, "parameter": self._parameter}
        rule_dict = {"name": "lowest_metric_variance", "n": n}
        self._append(params_dict, rule_dict)
        return self.create()

    def add_rule_most_frequent_best_dist(
        self,
        metric=None,
        parameter=None,
        n=1,
        dist_from_best_case=[0.01, 0.05, 0.1, 0.15],
    ):
        if metric is not None:
            self._metric = metric
        if parameter is not None:
            self._parameter = parameter
        params_dict = {"metric": self._metric, "parameter": self._parameter}
        rule_dict = {
            "name": "most_frequent_best_dist",
            "dist_from_best_case": dist_from_best_case,
            "n": n,
        }
        self._append(params_dict, rule_dict)
        return self.create()

    def add_rule_best_avg_recency_weight(
        self,
        metric=None,
        parameter=None,
        n=1,
        curr_weight=[1.5, 2.0, 5.0],
        decay_type=["linear"],
    ):
        if metric is not None:
            self._metric = metric
        if parameter is not None:
            self._parameter = parameter
        params_dict = {"metric": metric, "parameter": parameter}
        rule_dict = {
            "name": "best_avg_recency_weight",
            "curr_weight": curr_weight,
            "decay_type": decay_type,
            "n": n,
        }
        self._append(params_dict, rule_dict)
        return self.create()

    def add_rule_best_avg_var_penalized(
        self, metric=None, parameter=None, stdev_penalty=0.5, n=1
    ):
        if metric is not None:
            self._metric = metric
        if parameter is not None:
            self._parameter = parameter
        params_dict = {"metric": metric, "parameter": parameter}
        rule_dict = {
            "name": "best_avg_var_penalized",
            "stdev_penalty": stdev_penalty,
            "n": n,
        }
        self._append(params_dict, rule_dict)
        return self.create()


class RandomGroupRuleMaker(BaseRules):
    
    """
    The `RandomGroupRuleMaker` class generates a rule that randomly selects `n`
    model groups for each train set.

    Unlike the other two RuleMaker classes, it generates its selection rule spec
    on `__init__`
    """
    
    def __init__(self, n=1):
        self.shared_parameters = [{}]
        self.selection_rules = [{"name": "random_model_group", "n": n}]


class TwoMetricsRuleMaker(BaseRules):

    """
    The `TwoMetricsRuleMaker` class allows for the specification of rules that 
    evaluate a model group's performance in terms of two metrics. It currently
     supports one rule:

     - [best_average_two_metrics][triage.component.audition.selection_rules.best_average_two_metrics]

    """

    def add_rule_best_average_two_metrics(
        self,
        metric1="precision@",
        parameter1="100_abs",
        metric2="recall@",
        parameter2="300_abs",
        metric1_weight=[0.5],
        n=1,
    ):
        params_dict = {"metric1": metric1, "parameter1": parameter1}
        rule_dict = {
            "name": "best_average_two_metrics",
            "metric1_weight": metric1_weight,
            "metric2": [metric2],
            "parameter2": [parameter2],
            "n": n,
        }
        self._append(params_dict, rule_dict)


def create_selection_grid(*args):
    return list(map(lambda r: r.create()[0], args))



from triage.component.catwalk.metrics import fpr
from triage.component.catwalk.evaluation import ModelEvaluator


def test_metric_directionality():
    """All metrics must be wrapped using the @Metric decorator available
    in catwalk.metrics to provide an `greater_is_better` attribute which must
    be one of True or False.
    """
    for met in ModelEvaluator.available_metrics.values():
        assert hasattr(met, "greater_is_better")
        assert met.greater_is_better in (True, False)


def test_fpr():
    predictions_binary = [1, 1, 1, 0, 0, 0, 0, 0]
    labels = [1, 1, 0, 1, 0, 0, 0, 1]

    result = fpr([], predictions_binary, labels, [])
    # false positives = 1
    # total negatives = 4
    assert result == 0.25



import verboselogs, logging
logger = verboselogs.VerboseLogger(__name__)

from triage.util.structs import FeatureNameList


class FeatureGroup(dict):
    def __init__(self, *args, **kwargs):
        try:
            self._names = [kwargs.pop("name")]
        except KeyError:
            self._names = []
        try:
            features = kwargs.pop("features_by_table")
        except KeyError:
            features = {}
        super(FeatureGroup, self).__init__(*args, **features)

    @property
    def names(self):
        return self._names

    def update(self, other_group):
        super(FeatureGroup, self).update(other_group)
        self._names += other_group.names


def table_subsetter(config_item, table, features):
    "Return features matching a given table"
    if table == config_item:
        return features
    else:
        return []


def prefix_subsetter(config_item, table, features):
    "Return features matching a given prefix"
    return [feature for feature in features if feature.startswith(config_item)]

def all_subsetter(config_item, table, features):
    return features


class FeatureGroupCreator:
    """Divides a feature dictionary into groups based on given criteria"""

    subsetters = {
        "tables": table_subsetter,
        "prefix": prefix_subsetter,
        "all": all_subsetter
    }

    def __init__(self, definition):
        """
        Args:
            definition (dict) rules for generating feature groups
                Each key must correspond to a key in self.subsetters
                Each value (a list) must be understood by the subsetter
        """
        self.definition = definition

    def validate(self):
        """Validate the object's configuration

        """
        if not isinstance(self.definition, dict):
            raise ValueError("Feature Group Definition must be a dictionary")

        for subsetter_name, value in self.definition.items():
            if subsetter_name not in self.subsetters:
                raise ValueError(f"Unknown subsetter {subsetter_name} received")
            if not hasattr(value, "__iter__") or isinstance(value, (str, bytes)):
                raise ValueError(
                    "Each value in FeatureGroupCreator must be "
                    "iterable and not a string"
                )

    def subsets(self, feature_dictionary):
        """Generate subsets of a feature dict

        Args:
            feature_dictionary (dict) tables and the features contained in each

            The feature dictionary is meant to be keyed on source table. Example:

            {
                'feature_table_one': ['feature_one', feature_two'],
                'feature_table_two': ['feature_three', 'feature_four'],
            }

        Returns: (list) subsets of the feature dictionary, in the same
            table-based structure
        """
        logger.spam(
            f"Creating feature groups, using: {self.definition}, Master feature dictionary: {feature_dictionary}",
        )
        subsets = []
        for name, config in sorted(self.definition.items()):
            logger.spam(f"Parsing config grouping method {name}, items {config}")
            for config_item in config:
                subset = FeatureGroup(name=f"{name}: {config_item}")
                logger.spam(f"Finding columns that might belong in [{name}: {config_item}]")
                for table, features in feature_dictionary.items():
                    logger.spam(
                        f"Searching features in table {table} that match group {subset}"
                    )
                    matching_features = self.subsetters[name](
                        config_item, table, features
                    )
                    logger.debug(
                        f"Found {len(matching_features)} matching features in table {table} that match group [{name}: {config_item}]",
                    )
                    if len(matching_features) > 0:
                        subset[table] = FeatureNameList(matching_features)

                subsets.append(subset)

        if not any(subset for subset in subsets if any(subset)):
            raise ValueError(
                f"Problem! The feature group definition {self.definition} did not find any matches",
                f"in feature dictionary {feature_dictionary}"
            )
        logger.verbose(f"Found {len(subsets)} total feature subsets")
        return subsets
